import asyncio
import logging

from collections.abc import AsyncGenerator

from pydantic import ValidationError

from a2a.server.events.event_queue import Event, EventQueueLegacy, QueueShutDown
from a2a.types.a2a_pb2 import (
    Message,
    Task,
    TaskState,
    TaskStatusUpdateEvent,
)
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)


@trace_class(kind=SpanKind.SERVER)
class EventConsumer:
    """Consumer to read events from the agent event queue."""

    def __init__(self, queue: EventQueueLegacy):
        """Initializes the EventConsumer.

        Args:
            queue: The `EventQueue` instance to consume events from.
        """
        self.queue = queue
        self._timeout = 0.5
        self._exception: BaseException | None = None
        logger.debug('EventConsumer initialized')

    async def consume_all(self) -> AsyncGenerator[Event]:
        """Consume all the generated streaming events from the agent.

        This method yields events as they become available from the queue
        until a final event is received or the queue is closed. It also
        monitors for exceptions set by the `agent_task_callback`.

        Yields:
            Events dequeued from the queue.

        Raises:
            BaseException: If an exception was set by the `agent_task_callback`.
        """
        logger.debug('Starting to consume all events from the queue.')
        while True:
            if self._exception:
                raise self._exception
            try:
                # We use a timeout when waiting for an event from the queue.
                # This is required because it allows the loop to check if
                # `self._exception` has been set by the `agent_task_callback`.
                # Without the timeout, loop might hang indefinitely if no events are
                # enqueued by the agent and the agent simply threw an exception
                event = await asyncio.wait_for(
                    self.queue.dequeue_event(), timeout=self._timeout
                )
                logger.debug(
                    'Dequeued event of type: %s in consume_all.', type(event)
                )
                self.queue.task_done()
                logger.debug(
                    'Marked task as done in event queue in consume_all'
                )

                is_final_event = isinstance(event, Message) or (
                    isinstance(event, Task | TaskStatusUpdateEvent)
                    and event.status.state
                    in (
                        TaskState.TASK_STATE_COMPLETED,
                        TaskState.TASK_STATE_CANCELED,
                        TaskState.TASK_STATE_FAILED,
                        TaskState.TASK_STATE_REJECTED,
                        TaskState.TASK_STATE_UNSPECIFIED,
                        TaskState.TASK_STATE_INPUT_REQUIRED,
                    )
                )

                # Make sure the yield is after the close events, otherwise
                # the caller may end up in a blocked state where this
                # generator isn't called again to close things out and the
                # other part is waiting for an event or a closed queue.
                if is_final_event:
                    logger.debug('Stopping event consumption in consume_all.')
                    await self.queue.close(True)
                    yield event
                    break
                yield event
            except TimeoutError:
                # continue polling until there is a final event
                continue
            except asyncio.TimeoutError:  # pyright: ignore [reportUnusedExcept]
                # This class was made an alias of built-in TimeoutError after 3.11
                continue
            except (QueueShutDown, asyncio.QueueEmpty):
                # Confirm that the queue is closed, e.g. we aren't on
                # python 3.12 and get a queue empty error on an open queue
                if self.queue.is_closed():
                    break
            except ValidationError:
                logger.exception('Invalid event format received')
                continue
            except Exception as e:
                logger.exception('Stopping event consumption due to exception')
                self._exception = e
                continue

    def agent_task_callback(self, agent_task: asyncio.Task[None]) -> None:
        """Callback to handle exceptions from the agent's execution task.

        If the agent's asyncio task raises an exception, this callback is
        invoked, and the exception is stored to be re-raised by the consumer loop.

        Args:
            agent_task: The asyncio.Task that completed.
        """
        logger.debug('Agent task callback triggered.')
        if not agent_task.cancelled() and agent_task.done():
            self._exception = agent_task.exception()



import json
import logging

from collections.abc import AsyncIterator, Awaitable, Callable
from typing import TYPE_CHECKING, Any, TypeVar

from google.protobuf.json_format import MessageToDict, Parse

from a2a.server.context import ServerCallContext
from a2a.server.request_handlers.request_handler import RequestHandler
from a2a.server.routes.common import (
    DefaultServerCallContextBuilder,
    ServerCallContextBuilder,
)
from a2a.types import a2a_pb2
from a2a.types.a2a_pb2 import (
    CancelTaskRequest,
    GetTaskPushNotificationConfigRequest,
    SubscribeToTaskRequest,
)
from a2a.utils import constants, proto_utils
from a2a.utils.error_handlers import (
    build_rest_error_payload,
    rest_error_handler,
    rest_stream_error_handler,
)
from a2a.utils.errors import (
    InvalidRequestError,
    TaskNotFoundError,
)
from a2a.utils.telemetry import SpanKind, trace_class
from a2a.utils.version_validator import validate_version


if TYPE_CHECKING:
    from sse_starlette.event import ServerSentEvent
    from sse_starlette.sse import EventSourceResponse
    from starlette.requests import Request
    from starlette.responses import JSONResponse, Response

    _package_starlette_installed = True
else:
    try:
        from sse_starlette.event import ServerSentEvent
        from sse_starlette.sse import EventSourceResponse
        from starlette.requests import Request
        from starlette.responses import JSONResponse, Response

        _package_starlette_installed = True
    except ImportError:
        EventSourceResponse = Any
        ServerSentEvent = Any
        Request = Any
        JSONResponse = Any
        Response = Any

        _package_starlette_installed = False

logger = logging.getLogger(__name__)

TResponse = TypeVar('TResponse')


@trace_class(kind=SpanKind.SERVER)
class RestDispatcher:
    """Dispatches incoming REST requests to the appropriate handler methods.

    Handles context building, routing to RequestHandler directly, and response formatting (JSON/SSE).
    """

    def __init__(
        self,
        request_handler: RequestHandler,
        context_builder: ServerCallContextBuilder | None = None,
    ) -> None:
        """Initializes the RestDispatcher.

        Args:
            request_handler: The underlying `RequestHandler` instance to delegate requests to.
            context_builder: The ServerCallContextBuilder used to construct the
              ServerCallContext passed to the request_handler. If None the
              DefaultServerCallContextBuilder is used.
        """
        if not _package_starlette_installed:
            raise ImportError(
                'Packages `starlette` and `sse-starlette` are required to use the'
                ' `RestDispatcher`. They can be added as a part of `a2a-sdk` '
                'optional dependencies, `a2a-sdk[http-server]`.'
            )

        self._context_builder = (
            context_builder or DefaultServerCallContextBuilder()
        )
        self.request_handler = request_handler

    def _build_call_context(self, request: Request) -> ServerCallContext:
        call_context = self._context_builder.build(request)
        if 'tenant' in request.path_params:
            call_context.tenant = request.path_params['tenant']
        return call_context

    async def _handle_non_streaming(
        self,
        request: Request,
        handler_func: Callable[[ServerCallContext], Awaitable[TResponse]],
    ) -> TResponse:
        """Centralized error handling and context management for unary calls."""
        context = self._build_call_context(request)
        return await handler_func(context)

    async def _handle_streaming(
        self,
        request: Request,
        handler_func: Callable[[ServerCallContext], AsyncIterator[Any]],
    ) -> EventSourceResponse:
        """Centralized error handling and context management for streaming calls."""
        # Pre-consume and cache the request body to prevent deadlock in streaming context
        # This is required because Starlette's request.body() can only be consumed once,
        # and attempting to consume it after EventSourceResponse starts causes deadlock
        try:
            await request.body()
        except (ValueError, RuntimeError, OSError) as e:
            raise InvalidRequestError(
                message=f'Failed to pre-consume request body: {e}'
            ) from e

        context = self._build_call_context(request)

        # Eagerly fetch the first item from the stream so that errors raised
        # before any event is yielded (e.g. validation, parsing, or handler
        # failures) propagate here and are caught by
        # @rest_stream_error_handler, which returns a JSONResponse with
        # the correct HTTP status code instead of starting an SSE stream.
        # Without this, the error would be raised after SSE headers are
        # already sent, and the client would see a broken stream instead
        stream = aiter(handler_func(context))
        try:
            first_item = await anext(stream)
        except StopAsyncIteration:
            return EventSourceResponse(iter([]))

        async def event_generator() -> AsyncIterator[ServerSentEvent]:
            yield ServerSentEvent(data=json.dumps(first_item))
            try:
                async for item in stream:
                    yield ServerSentEvent(data=json.dumps(item))
            except Exception as e:
                logger.exception('Error during REST SSE stream')
                yield ServerSentEvent(
                    data=json.dumps(build_rest_error_payload(e)),
                    event='error',
                )

        return EventSourceResponse(event_generator())

    @rest_error_handler
    async def on_message_send(self, request: Request) -> Response:
        """Handles the 'message/send' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> a2a_pb2.SendMessageResponse:
            body = await request.body()
            params = a2a_pb2.SendMessageRequest()
            Parse(body, params)
            task_or_message = await self.request_handler.on_message_send(
                params, context
            )
            if isinstance(task_or_message, a2a_pb2.Task):
                return a2a_pb2.SendMessageResponse(task=task_or_message)
            return a2a_pb2.SendMessageResponse(message=task_or_message)

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))

    @rest_stream_error_handler
    async def on_message_send_stream(
        self, request: Request
    ) -> EventSourceResponse:
        """Handles the 'message/stream' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> AsyncIterator[dict[str, Any]]:
            body = await request.body()
            params = a2a_pb2.SendMessageRequest()
            Parse(body, params)
            async for event in self.request_handler.on_message_send_stream(
                params, context
            ):
                response = proto_utils.to_stream_response(event)
                yield MessageToDict(response)

        return await self._handle_streaming(request, _handler)

    @rest_error_handler
    async def on_cancel_task(self, request: Request) -> Response:
        """Handles the 'tasks/cancel' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(context: ServerCallContext) -> a2a_pb2.Task:
            task_id = request.path_params['id']
            task = await self.request_handler.on_cancel_task(
                CancelTaskRequest(id=task_id), context
            )
            if task:
                return task
            raise TaskNotFoundError

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))

    @rest_stream_error_handler
    async def on_subscribe_to_task(
        self, request: Request
    ) -> EventSourceResponse:
        """Handles the 'SubscribeToTask' REST method."""
        task_id = request.path_params['id']

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> AsyncIterator[dict[str, Any]]:
            async for event in self.request_handler.on_subscribe_to_task(
                SubscribeToTaskRequest(id=task_id), context
            ):
                response = proto_utils.to_stream_response(event)
                yield MessageToDict(response)

        return await self._handle_streaming(request, _handler)

    @rest_error_handler
    async def on_get_task(self, request: Request) -> Response:
        """Handles the 'tasks/{id}' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(context: ServerCallContext) -> a2a_pb2.Task:
            params = a2a_pb2.GetTaskRequest()
            proto_utils.parse_params(request.query_params, params)
            params.id = request.path_params['id']
            task = await self.request_handler.on_get_task(params, context)
            if task:
                return task
            raise TaskNotFoundError

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))

    @rest_error_handler
    async def get_push_notification(self, request: Request) -> Response:
        """Handles the 'tasks/pushNotificationConfig/get' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> a2a_pb2.TaskPushNotificationConfig:
            task_id = request.path_params['id']
            push_id = request.path_params['push_id']
            params = GetTaskPushNotificationConfigRequest(
                task_id=task_id, id=push_id
            )
            return (
                await self.request_handler.on_get_task_push_notification_config(
                    params, context
                )
            )

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))

    @rest_error_handler
    async def delete_push_notification(self, request: Request) -> Response:
        """Handles the 'tasks/pushNotificationConfig/delete' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(context: ServerCallContext) -> None:
            task_id = request.path_params['id']
            push_id = request.path_params['push_id']
            params = a2a_pb2.DeleteTaskPushNotificationConfigRequest(
                task_id=task_id, id=push_id
            )
            await self.request_handler.on_delete_task_push_notification_config(
                params, context
            )

        await self._handle_non_streaming(request, _handler)
        return JSONResponse(content={})

    @rest_error_handler
    async def set_push_notification(self, request: Request) -> Response:
        """Handles the 'tasks/pushNotificationConfig/set' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> a2a_pb2.TaskPushNotificationConfig:
            body = await request.body()
            params = a2a_pb2.TaskPushNotificationConfig()
            Parse(body, params)
            params.task_id = request.path_params['id']
            return await self.request_handler.on_create_task_push_notification_config(
                params, context
            )

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))

    @rest_error_handler
    async def list_push_notifications(self, request: Request) -> Response:
        """Handles the 'tasks/pushNotificationConfig/list' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> a2a_pb2.ListTaskPushNotificationConfigsResponse:
            params = a2a_pb2.ListTaskPushNotificationConfigsRequest()
            proto_utils.parse_params(request.query_params, params)
            params.task_id = request.path_params['id']
            return await self.request_handler.on_list_task_push_notification_configs(
                params, context
            )

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))

    @rest_error_handler
    async def list_tasks(self, request: Request) -> Response:
        """Handles the 'tasks/list' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> a2a_pb2.ListTasksResponse:
            params = a2a_pb2.ListTasksRequest()
            proto_utils.parse_params(request.query_params, params)
            return await self.request_handler.on_list_tasks(params, context)

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(
            content=MessageToDict(
                response, always_print_fields_with_no_presence=True
            )
        )

    @rest_error_handler
    async def handle_authenticated_agent_card(
        self, request: Request
    ) -> Response:
        """Handles the 'agentCard' REST method."""

        @validate_version(constants.PROTOCOL_VERSION_1_0)
        async def _handler(
            context: ServerCallContext,
        ) -> a2a_pb2.AgentCard:
            params = a2a_pb2.GetExtendedAgentCardRequest()
            return await self.request_handler.on_get_extended_agent_card(
                params, context
            )

        response = await self._handle_non_streaming(request, _handler)
        return JSONResponse(content=MessageToDict(response))



import functools
import inspect
import logging

from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from starlette.responses import JSONResponse, Response
else:
    try:
        from starlette.responses import JSONResponse, Response
    except ImportError:
        JSONResponse = Any
        Response = Any


from google.protobuf.json_format import ParseError

from a2a.utils.errors import (
    A2A_REST_ERROR_MAPPING,
    A2AError,
    InternalError,
    RestErrorMap,
)


logger = logging.getLogger(__name__)


def _build_error_payload(
    code: int,
    status: str,
    message: str,
    reason: str | None = None,
    metadata: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Helper function to build the JSON error payload."""
    payload: dict[str, Any] = {
        'code': code,
        'status': status,
        'message': message,
    }
    if reason:
        payload['details'] = [
            {
                '@type': 'type.googleapis.com/google.rpc.ErrorInfo',
                'reason': reason,
                'domain': 'a2a-protocol.org',
                'metadata': metadata if metadata is not None else {},
            }
        ]
    return {'error': payload}


def build_rest_error_payload(error: Exception) -> dict[str, Any]:
    """Build a REST error payload dict from an exception.

    Returns:
        A dict with the error payload in the standard REST error format.
    """
    if isinstance(error, A2AError):
        mapping = A2A_REST_ERROR_MAPPING.get(
            type(error), RestErrorMap(500, 'INTERNAL', 'INTERNAL_ERROR')
        )
        # SECURITY WARNING: Data attached to A2AError.data is serialized unaltered and exposed publicly to the client in the REST API response.
        metadata = getattr(error, 'data', None) or {}
        return _build_error_payload(
            code=mapping.http_code,
            status=mapping.grpc_status,
            message=getattr(error, 'message', str(error)),
            reason=mapping.reason,
            metadata=metadata,
        )
    if isinstance(error, ParseError):
        return _build_error_payload(
            code=400,
            status='INVALID_ARGUMENT',
            message=str(error),
            reason='INVALID_REQUEST',
            metadata={},
        )
    return _build_error_payload(
        code=500,
        status='INTERNAL',
        message='unknown exception',
    )


def _create_error_response(error: Exception) -> Response:
    """Helper function to create a JSONResponse for an error."""
    if isinstance(error, A2AError):
        log_level = (
            logging.ERROR
            if isinstance(error, InternalError)
            else logging.WARNING
        )
        logger.log(
            log_level,
            "Request error: Code=%s, Message='%s'%s",
            getattr(error, 'code', 'N/A'),
            getattr(error, 'message', str(error)),
            f', Data={error.data}' if error.data else '',
        )
    elif isinstance(error, ParseError):
        logger.warning('Parse error: %s', str(error))
    else:
        logger.exception('Unknown error occurred')

    payload = build_rest_error_payload(error)
    # Extract HTTP status code from the payload
    http_code = payload.get('error', {}).get('code', 500)
    return JSONResponse(
        content=payload,
        status_code=http_code,
        media_type='application/json',
    )


def rest_error_handler(
    func: Callable[..., Awaitable[Response]],
) -> Callable[..., Awaitable[Response]]:
    """Decorator to catch A2AError and map it to an appropriate JSONResponse."""

    @functools.wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Response:
        try:
            return await func(*args, **kwargs)
        except Exception as error:  # noqa: BLE001
            return _create_error_response(error)

    return wrapper


def rest_stream_error_handler(
    func: Callable[..., Coroutine[Any, Any, Any]],
) -> Callable[..., Coroutine[Any, Any, Any]]:
    """Decorator to catch A2AError for a streaming method. Maps synchronous errors to JSONResponse and logs streaming errors."""

    def _log_error(error: Exception) -> None:
        if isinstance(error, A2AError):
            log_level = (
                logging.ERROR
                if isinstance(error, InternalError)
                else logging.WARNING
            )
            logger.log(
                log_level,
                "Request error: Code=%s, Message='%s'%s",
                getattr(error, 'code', 'N/A'),
                getattr(error, 'message', str(error)),
                f', Data={error.data}' if error.data else '',
            )
        else:
            logger.exception('Unknown streaming error occurred')

    @functools.wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            response = await func(*args, **kwargs)

            # If the response has an async generator body (like EventSourceResponse),
            # we must wrap it to catch errors that occur during stream execution.
            if hasattr(response, 'body_iterator') and inspect.isasyncgen(
                response.body_iterator
            ):
                original_iterator = response.body_iterator

                async def error_catching_iterator() -> AsyncGenerator[
                    Any, None
                ]:
                    try:
                        async for item in original_iterator:
                            yield item
                    except Exception as stream_error:
                        _log_error(stream_error)
                        raise stream_error

                response.body_iterator = error_catching_iterator()

        except Exception as e:  # noqa: BLE001
            return _create_error_response(e)
        else:
            return response

    return wrapper



"""Unified helper functions for creating and handling A2A types."""

import uuid

from collections.abc import Sequence
from typing import Any

from google.protobuf import struct_pb2
from google.protobuf.json_format import ParseDict

from a2a.types.a2a_pb2 import (
    Artifact,
    Message,
    Part,
    Role,
    StreamResponse,
    Task,
    TaskArtifactUpdateEvent,
    TaskState,
    TaskStatus,
    TaskStatusUpdateEvent,
)


# --- Message Helpers ---


def new_message(
    parts: list[Part],
    context_id: str | None = None,
    task_id: str | None = None,
    role: Role = Role.ROLE_AGENT,
) -> Message:
    """Creates a new message containing a list of Parts."""
    return Message(
        role=role,
        parts=parts,
        message_id=str(uuid.uuid4()),
        task_id=task_id,
        context_id=context_id,
    )


def new_text_message(
    text: str,
    media_type: str | None = None,
    context_id: str | None = None,
    task_id: str | None = None,
    role: Role = Role.ROLE_AGENT,
) -> Message:
    """Creates a new message containing a single text Part."""
    return new_message(
        parts=[new_text_part(text, media_type=media_type)],
        context_id=context_id,
        task_id=task_id,
        role=role,
    )


def get_message_text(message: Message, delimiter: str = '\n') -> str:
    """Extracts and joins all text content from a Message's parts."""
    return delimiter.join(get_text_parts(message.parts))


def new_data_message(
    data: Any,
    media_type: str | None = None,
    context_id: str | None = None,
    task_id: str | None = None,
    role: Role = Role.ROLE_AGENT,
) -> Message:
    """Creates a new message containing a single data Part.

    Args:
        data: JSON-serializable data to embed (dict, list, str, etc.).
        media_type: Optional MIME type of the part content (e.g., "text/plain", "application/json", "image/png").
        context_id: Optional context ID.
        task_id: Optional task ID.
        role: The role of the message sender (default: ROLE_AGENT).

    Returns:
        A Message with a single data Part.
    """
    return new_message(
        parts=[new_data_part(data, media_type=media_type)],
        context_id=context_id,
        task_id=task_id,
        role=role,
    )


def new_raw_message(  # noqa: PLR0913
    raw: bytes,
    media_type: str | None = None,
    filename: str | None = None,
    context_id: str | None = None,
    task_id: str | None = None,
    role: Role = Role.ROLE_AGENT,
) -> Message:
    """Creates a new message containing a single raw bytes Part.

    Args:
        raw: The raw bytes content.
        media_type: Optional MIME type (e.g. 'image/png').
        filename: Optional filename.
        context_id: Optional context ID.
        task_id: Optional task ID.
        role: The role of the message sender (default: ROLE_AGENT).

    Returns:
        A Message with a single raw Part.
    """
    return new_message(
        parts=[new_raw_part(raw, media_type=media_type, filename=filename)],
        context_id=context_id,
        task_id=task_id,
        role=role,
    )


def new_url_message(  # noqa: PLR0913
    url: str,
    media_type: str | None = None,
    filename: str | None = None,
    context_id: str | None = None,
    task_id: str | None = None,
    role: Role = Role.ROLE_AGENT,
) -> Message:
    """Creates a new message containing a single URL Part.

    Args:
        url: The URL pointing to the file content.
        media_type: Optional MIME type (e.g. 'image/png').
        filename: Optional filename.
        context_id: Optional context ID.
        task_id: Optional task ID.
        role: The role of the message sender (default: ROLE_AGENT).

    Returns:
        A Message with a single URL Part.
    """
    return new_message(
        parts=[new_url_part(url, media_type=media_type, filename=filename)],
        context_id=context_id,
        task_id=task_id,
        role=role,
    )


# --- Artifact Helpers ---


def new_artifact(
    parts: list[Part],
    name: str,
    description: str | None = None,
    artifact_id: str | None = None,
) -> Artifact:
    """Creates a new Artifact object."""
    return Artifact(
        artifact_id=artifact_id or str(uuid.uuid4()),
        parts=parts,
        name=name,
        description=description,
    )


def new_text_artifact(
    name: str,
    text: str,
    media_type: str | None = None,
    description: str | None = None,
    artifact_id: str | None = None,
) -> Artifact:
    """Creates a new Artifact object containing only a single text Part."""
    return new_artifact(
        [new_text_part(text, media_type=media_type)],
        name,
        description,
        artifact_id=artifact_id,
    )


def new_data_artifact(
    name: str,
    data: Any,
    media_type: str | None = None,
    description: str | None = None,
    artifact_id: str | None = None,
) -> Artifact:
    """Creates a new Artifact object containing only a single data Part.

    Args:
        name: The name of the artifact.
        data: JSON-serializable data to embed (dict, list, str, etc.).
        media_type: Optional MIME type of the part content (e.g., "text/plain", "application/json", "image/png").
        description: Optional description.
        artifact_id: Optional artifact ID (auto-generated if not provided).

    Returns:
        An Artifact with a single data Part.
    """
    return new_artifact(
        [new_data_part(data, media_type=media_type)],
        name,
        description,
        artifact_id=artifact_id,
    )


def new_raw_artifact(  # noqa: PLR0913
    name: str,
    raw: bytes,
    media_type: str | None = None,
    filename: str | None = None,
    description: str | None = None,
    artifact_id: str | None = None,
) -> Artifact:
    """Creates a new Artifact object containing only a single raw bytes Part.

    Args:
        name: The name of the artifact.
        raw: The raw bytes content.
        media_type: Optional MIME type (e.g. 'image/png').
        filename: Optional filename.
        description: Optional description.
        artifact_id: Optional artifact ID (auto-generated if not provided).

    Returns:
        An Artifact with a single raw Part.
    """
    return new_artifact(
        [new_raw_part(raw, media_type=media_type, filename=filename)],
        name,
        description,
        artifact_id=artifact_id,
    )


def new_url_artifact(  # noqa: PLR0913
    name: str,
    url: str,
    media_type: str | None = None,
    filename: str | None = None,
    description: str | None = None,
    artifact_id: str | None = None,
) -> Artifact:
    """Creates a new Artifact object containing only a single URL Part.

    Args:
        name: The name of the artifact.
        url: The URL pointing to the file content.
        media_type: Optional MIME type (e.g. 'image/png').
        filename: Optional filename.
        description: Optional description.
        artifact_id: Optional artifact ID (auto-generated if not provided).

    Returns:
        An Artifact with a single URL Part.
    """
    return new_artifact(
        [new_url_part(url, media_type=media_type, filename=filename)],
        name,
        description,
        artifact_id=artifact_id,
    )


def get_artifact_text(artifact: Artifact, delimiter: str = '\n') -> str:
    """Extracts and joins all text content from an Artifact's parts."""
    return delimiter.join(get_text_parts(artifact.parts))


# --- Task Helpers ---


def new_task_from_user_message(user_message: Message) -> Task:
    """Creates a new Task object from an initial user message."""
    if user_message.role != Role.ROLE_USER:
        raise ValueError('Message must be from a user')
    if not user_message.parts:
        raise ValueError('Message parts cannot be empty')
    for part in user_message.parts:
        if part.HasField('text') and not part.text:
            raise ValueError('Message.text cannot be empty')

    return Task(
        status=TaskStatus(state=TaskState.TASK_STATE_SUBMITTED),
        id=user_message.task_id or str(uuid.uuid4()),
        context_id=user_message.context_id or str(uuid.uuid4()),
        history=[user_message],
    )


def new_task(
    task_id: str,
    context_id: str,
    state: TaskState,
    artifacts: list[Artifact] | None = None,
    history: list[Message] | None = None,
) -> Task:
    """Creates a Task object with a specified status."""
    if history is None:
        history = []
    if artifacts is None:
        artifacts = []

    return Task(
        status=TaskStatus(state=state),
        id=task_id,
        context_id=context_id,
        artifacts=artifacts,
        history=history,
    )


# --- Part Helpers ---


def new_text_part(
    text: str,
    media_type: str | None = None,
) -> Part:
    """Creates a Part with text content.

    Args:
        text: The text content.
        media_type: Optional MIME type (e.g. 'text/plain', 'text/markdown').

    Returns:
        A Part with the text field set.
    """
    return Part(text=text, media_type=media_type or '')


def new_data_part(
    data: Any,
    media_type: str | None = None,
) -> Part:
    """Creates a Part with structured data (google.protobuf.Value).

    Args:
        data: JSON-serializable data to embed (dict, list, str, etc.).
        media_type: Optional MIME type of the part content (e.g., "text/plain", "application/json", "image/png").

    Returns:
        A Part with the data field set.
    """
    return Part(
        data=ParseDict(data, struct_pb2.Value()),
        media_type=media_type or '',
    )


def new_raw_part(
    raw: bytes,
    media_type: str | None = None,
    filename: str | None = None,
) -> Part:
    """Creates a Part with raw bytes content.

    Args:
        raw: The raw bytes content.
        media_type: Optional MIME type (e.g. 'image/png').
        filename: Optional filename.

    Returns:
        A Part with the raw field set.
    """
    return Part(
        raw=raw,
        media_type=media_type or '',
        filename=filename or '',
    )


def new_url_part(
    url: str,
    media_type: str | None = None,
    filename: str | None = None,
) -> Part:
    """Creates a Part with a URL pointing to file content.

    Args:
        url: The URL to the file content.
        media_type: Optional MIME type (e.g. 'image/png').
        filename: Optional filename.

    Returns:
        A Part with the url field set.
    """
    return Part(
        url=url,
        media_type=media_type or '',
        filename=filename or '',
    )


def get_text_parts(parts: Sequence[Part]) -> list[str]:
    """Extracts text content from all text Parts."""
    return [part.text for part in parts if part.HasField('text')]


# --- Event & Stream Helpers ---


def new_text_status_update_event(
    task_id: str,
    context_id: str,
    state: TaskState,
    text: str,
) -> TaskStatusUpdateEvent:
    """Creates a TaskStatusUpdateEvent with a single text message."""
    return TaskStatusUpdateEvent(
        task_id=task_id,
        context_id=context_id,
        status=TaskStatus(
            state=state,
            message=new_text_message(
                text=text,
                role=Role.ROLE_AGENT,
                context_id=context_id,
                task_id=task_id,
            ),
        ),
    )


def new_text_artifact_update_event(  # noqa: PLR0913
    task_id: str,
    context_id: str,
    name: str,
    text: str,
    append: bool = False,
    last_chunk: bool = False,
    artifact_id: str | None = None,
) -> TaskArtifactUpdateEvent:
    """Creates a TaskArtifactUpdateEvent with a single text artifact."""
    return TaskArtifactUpdateEvent(
        task_id=task_id,
        context_id=context_id,
        artifact=new_text_artifact(
            name=name, text=text, artifact_id=artifact_id
        ),
        append=append,
        last_chunk=last_chunk,
    )


def get_stream_response_text(
    response: StreamResponse, delimiter: str = '\n'
) -> str:
    """Extracts text content from a StreamResponse."""
    if response.HasField('message'):
        return get_message_text(response.message, delimiter)
    if response.HasField('task'):
        texts = [
            get_artifact_text(a, delimiter) for a in response.task.artifacts
        ]
        return delimiter.join(t for t in texts if t)
    if response.HasField('status_update'):
        if response.status_update.status.HasField('message'):
            return get_message_text(
                response.status_update.status.message, delimiter
            )
        return ''
    if response.HasField('artifact_update'):
        return get_artifact_text(response.artifact_update.artifact, delimiter)
    return ''



import json

from collections.abc import AsyncGenerator, Callable, Iterator
from contextlib import contextmanager
from typing import Any, NoReturn

import httpx

from httpx_sse import EventSource, SSEError

from a2a.client.client import ClientCallContext
from a2a.client.errors import A2AClientError, A2AClientTimeoutError


def _default_sse_error_handler(sse_data: str) -> NoReturn:
    raise A2AClientError(f'SSE stream error event received: {sse_data}')


@contextmanager
def handle_http_exceptions(
    status_error_handler: Callable[[httpx.HTTPStatusError], NoReturn]
    | None = None,
) -> Iterator[None]:
    """Handles common HTTP exceptions for REST and JSON-RPC transports.

    Args:
        status_error_handler: Optional handler for `httpx.HTTPStatusError`.
            If provided, this handler should raise an appropriate domain-specific exception.
            If not provided, a default `A2AClientError` will be raised.
    """
    try:
        yield
    except httpx.TimeoutException as e:
        raise A2AClientTimeoutError('Client Request timed out') from e
    except httpx.HTTPStatusError as e:
        if status_error_handler:
            status_error_handler(e)
        raise A2AClientError(f'HTTP Error {e.response.status_code}: {e}') from e
    except SSEError as e:
        raise A2AClientError(
            f'Invalid SSE response or protocol error: {e}'
        ) from e
    except httpx.RequestError as e:
        raise A2AClientError(f'Network communication error: {e}') from e
    except json.JSONDecodeError as e:
        raise A2AClientError(f'JSON Decode Error: {e}') from e


def get_http_args(context: ClientCallContext | None) -> dict[str, Any]:
    """Extracts HTTP arguments from the client call context."""
    http_kwargs: dict[str, Any] = {}
    if context and context.service_parameters:
        http_kwargs['headers'] = context.service_parameters.copy()
    if context and context.timeout is not None:
        http_kwargs['timeout'] = httpx.Timeout(context.timeout)
    return http_kwargs


async def send_http_request(
    httpx_client: httpx.AsyncClient,
    request: httpx.Request,
    status_error_handler: Callable[[httpx.HTTPStatusError], NoReturn]
    | None = None,
) -> dict[str, Any]:
    """Sends an HTTP request and parses the JSON response, handling common exceptions."""
    with handle_http_exceptions(status_error_handler):
        response = await httpx_client.send(request)
        response.raise_for_status()
        return response.json()


async def send_http_stream_request(
    httpx_client: httpx.AsyncClient,
    method: str,
    url: str,
    status_error_handler: Callable[[httpx.HTTPStatusError], NoReturn]
    | None = None,
    sse_error_handler: Callable[[str], NoReturn] = _default_sse_error_handler,
    **kwargs: Any,
) -> AsyncGenerator[str]:
    """Sends a streaming HTTP request, yielding SSE data strings and handling exceptions.

    Args:
        httpx_client: The async HTTP client.
        method: The HTTP method (e.g. 'POST', 'GET').
        url: The URL to send the request to.
        status_error_handler: Handler for HTTP status errors. Should raise an
            appropriate domain-specific exception.
        sse_error_handler: Handler for SSE error events. Called with the
            raw SSE data string when an ``event: error`` SSE event is received.
            Should raise an appropriate domain-specific exception.
        **kwargs: Additional keyword arguments forwarded to ``aconnect_sse``.
    """
    with handle_http_exceptions(status_error_handler):
        async with _SSEEventSource(
            httpx_client, method, url, **kwargs
        ) as event_source:
            try:
                event_source.response.raise_for_status()
            except httpx.HTTPStatusError as e:
                # Read upfront streaming error content immediately, otherwise lower-level handlers
                # (e.g. response.json()) crash with 'ResponseNotRead' Access errors.
                await event_source.response.aread()
                raise e

            # If the response is not a stream, read it standardly (e.g., upfront JSON-RPC error payload)
            if 'text/event-stream' not in event_source.response.headers.get(
                'content-type', ''
            ):
                content = await event_source.response.aread()
                yield content.decode('utf-8')
                return

            async for sse in event_source.aiter_sse():
                if not sse.data:
                    continue
                if sse.event == 'error':
                    sse_error_handler(sse.data)
                yield sse.data


class _SSEEventSource:
    """Class-based replacement for ``httpx_sse.aconnect_sse``.

    ``aconnect_sse`` is an ``@asynccontextmanager`` whose internal async
    generator gets tracked by the event loop. When the enclosing async
    generator is abandoned, the event loop's generator cleanup collides
    with the cascading cleanup — see https://bugs.python.org/issue38559.

    Plain ``__aenter__``/``__aexit__`` coroutines avoid this entirely.
    """

    def __init__(
        self,
        client: httpx.AsyncClient,
        method: str,
        url: str,
        **kwargs: Any,
    ) -> None:
        headers = httpx.Headers(kwargs.pop('headers', None))
        headers.setdefault('Accept', 'text/event-stream')
        headers.setdefault('Cache-Control', 'no-store')
        self._request = client.build_request(
            method, url, headers=headers, **kwargs
        )
        self._client = client
        self._response: httpx.Response | None = None

    async def __aenter__(self) -> EventSource:
        self._response = await self._client.send(self._request, stream=True)
        return EventSource(self._response)

    async def __aexit__(self, *args: object) -> None:
        if self._response is not None:
            await self._response.aclose()



# mypy: disable-error-code="arg-type"
"""This file was migrated from the a2a-python SDK version 0.3.

It provides utilities for converting between legacy v0.3 Pydantic models and legacy v0.3 Protobuf definitions.
"""

import json
import logging
import re

from typing import Any

from google.protobuf import json_format, struct_pb2

from a2a.compat.v0_3 import a2a_v0_3_pb2 as a2a_pb2
from a2a.compat.v0_3 import types
from a2a.utils.errors import InvalidParamsError


logger = logging.getLogger(__name__)


# Regexp patterns for matching
TASK_NAME_MATCH = re.compile(r'tasks/([^/]+)')
TASK_PUSH_CONFIG_NAME_MATCH = re.compile(
    r'tasks/([^/]+)/pushNotificationConfigs/([^/]+)'
)


def dict_to_struct(dictionary: dict[str, Any]) -> struct_pb2.Struct:
    """Converts a Python dict to a Struct proto.

    Unfortunately, using `json_format.ParseDict` does not work because this
    wants the dictionary to be an exact match of the Struct proto with fields
    and keys and values, not the traditional Python dict structure.

    Args:
      dictionary: The Python dict to convert.

    Returns:
      The Struct proto.
    """
    struct = struct_pb2.Struct()
    for key, val in dictionary.items():
        if isinstance(val, dict):
            struct[key] = dict_to_struct(val)
        else:
            struct[key] = val
    return struct


def make_dict_serializable(value: Any) -> Any:
    """Dict pre-processing utility: converts non-serializable values to serializable form.

    Use this when you want to normalize a dictionary before dict->Struct conversion.

    Args:
        value: The value to convert.

    Returns:
        A serializable value.
    """
    if isinstance(value, str | int | float | bool) or value is None:
        return value
    if isinstance(value, dict):
        return {k: make_dict_serializable(v) for k, v in value.items()}
    if isinstance(value, list | tuple):
        return [make_dict_serializable(item) for item in value]
    return str(value)


def normalize_large_integers_to_strings(
    value: Any, max_safe_digits: int = 15
) -> Any:
    """Integer preprocessing utility: converts large integers to strings.

    Use this when you want to convert large integers to strings considering
    JavaScript's MAX_SAFE_INTEGER (2^53 - 1) limitation.

    Args:
        value: The value to convert.
        max_safe_digits: Maximum safe integer digits (default: 15).

    Returns:
        A normalized value.
    """
    max_safe_int = 10**max_safe_digits - 1

    def _normalize(item: Any) -> Any:
        if isinstance(item, int) and abs(item) > max_safe_int:
            return str(item)
        if isinstance(item, dict):
            return {k: _normalize(v) for k, v in item.items()}
        if isinstance(item, list | tuple):
            return [_normalize(i) for i in item]
        return item

    return _normalize(value)


def parse_string_integers_in_dict(value: Any, max_safe_digits: int = 15) -> Any:
    """String post-processing utility: converts large integer strings back to integers.

    Use this when you want to restore large integer strings to integers
    after Struct->dict conversion.

    Args:
        value: The value to convert.
        max_safe_digits: Maximum safe integer digits (default: 15).

    Returns:
        A parsed value.
    """
    if isinstance(value, dict):
        return {
            k: parse_string_integers_in_dict(v, max_safe_digits)
            for k, v in value.items()
        }
    if isinstance(value, list | tuple):
        return [
            parse_string_integers_in_dict(item, max_safe_digits)
            for item in value
        ]
    if isinstance(value, str):
        # Handle potential negative numbers.
        stripped_value = value.lstrip('-')
        if stripped_value.isdigit() and len(stripped_value) > max_safe_digits:
            return int(value)
    return value


class ToProto:
    """Converts Python types to proto types."""

    @classmethod
    def message(cls, message: types.Message | None) -> a2a_pb2.Message | None:
        if message is None:
            return None
        return a2a_pb2.Message(
            message_id=message.message_id,
            content=[cls.part(p) for p in message.parts],
            context_id=message.context_id or '',
            task_id=message.task_id or '',
            role=cls.role(message.role),
            metadata=cls.metadata(message.metadata),
            extensions=message.extensions or [],
        )

    @classmethod
    def metadata(
        cls, metadata: dict[str, Any] | None
    ) -> struct_pb2.Struct | None:
        if metadata is None:
            return None
        return dict_to_struct(metadata)

    @classmethod
    def part(cls, part: types.Part) -> a2a_pb2.Part:
        if isinstance(part.root, types.TextPart):
            return a2a_pb2.Part(
                text=part.root.text, metadata=cls.metadata(part.root.metadata)
            )
        if isinstance(part.root, types.FilePart):
            return a2a_pb2.Part(
                file=cls.file(part.root.file),
                metadata=cls.metadata(part.root.metadata),
            )
        if isinstance(part.root, types.DataPart):
            return a2a_pb2.Part(
                data=cls.data(part.root.data),
                metadata=cls.metadata(part.root.metadata),
            )
        raise ValueError(f'Unsupported part type: {part.root}')

    @classmethod
    def data(cls, data: dict[str, Any]) -> a2a_pb2.DataPart:
        return a2a_pb2.DataPart(data=dict_to_struct(data))

    @classmethod
    def file(
        cls, file: types.FileWithUri | types.FileWithBytes
    ) -> a2a_pb2.FilePart:
        if isinstance(file, types.FileWithUri):
            return a2a_pb2.FilePart(
                file_with_uri=file.uri, mime_type=file.mime_type, name=file.name
            )
        return a2a_pb2.FilePart(
            file_with_bytes=file.bytes.encode('utf-8'),
            mime_type=file.mime_type,
            name=file.name,
        )

    @classmethod
    def task(cls, task: types.Task) -> a2a_pb2.Task:
        return a2a_pb2.Task(
            id=task.id,
            context_id=task.context_id,
            status=cls.task_status(task.status),
            artifacts=(
                [cls.artifact(a) for a in task.artifacts]
                if task.artifacts
                else None
            ),
            history=(
                [cls.message(h) for h in task.history]  # type: ignore[misc]
                if task.history
                else None
            ),
            metadata=cls.metadata(task.metadata),
        )

    @classmethod
    def task_status(cls, status: types.TaskStatus) -> a2a_pb2.TaskStatus:
        return a2a_pb2.TaskStatus(
            state=cls.task_state(status.state),
            update=cls.message(status.message),
        )

    @classmethod
    def task_state(cls, state: types.TaskState) -> a2a_pb2.TaskState:
        match state:
            case types.TaskState.submitted:
                return a2a_pb2.TaskState.TASK_STATE_SUBMITTED
            case types.TaskState.working:
                return a2a_pb2.TaskState.TASK_STATE_WORKING
            case types.TaskState.completed:
                return a2a_pb2.TaskState.TASK_STATE_COMPLETED
            case types.TaskState.canceled:
                return a2a_pb2.TaskState.TASK_STATE_CANCELLED
            case types.TaskState.failed:
                return a2a_pb2.TaskState.TASK_STATE_FAILED
            case types.TaskState.input_required:
                return a2a_pb2.TaskState.TASK_STATE_INPUT_REQUIRED
            case types.TaskState.auth_required:
                return a2a_pb2.TaskState.TASK_STATE_AUTH_REQUIRED
            case types.TaskState.rejected:
                return a2a_pb2.TaskState.TASK_STATE_REJECTED
            case _:
                return a2a_pb2.TaskState.TASK_STATE_UNSPECIFIED

    @classmethod
    def artifact(cls, artifact: types.Artifact) -> a2a_pb2.Artifact:
        return a2a_pb2.Artifact(
            artifact_id=artifact.artifact_id,
            description=artifact.description,
            metadata=cls.metadata(artifact.metadata),
            name=artifact.name,
            parts=[cls.part(p) for p in artifact.parts],
            extensions=artifact.extensions or [],
        )

    @classmethod
    def authentication_info(
        cls, info: types.PushNotificationAuthenticationInfo
    ) -> a2a_pb2.AuthenticationInfo:
        return a2a_pb2.AuthenticationInfo(
            schemes=info.schemes,
            credentials=info.credentials,
        )

    @classmethod
    def push_notification_config(
        cls, config: types.PushNotificationConfig
    ) -> a2a_pb2.PushNotificationConfig:
        auth_info = (
            cls.authentication_info(config.authentication)
            if config.authentication
            else None
        )
        return a2a_pb2.PushNotificationConfig(
            id=config.id or '',
            url=config.url,
            token=config.token,
            authentication=auth_info,
        )

    @classmethod
    def task_artifact_update_event(
        cls, event: types.TaskArtifactUpdateEvent
    ) -> a2a_pb2.TaskArtifactUpdateEvent:
        return a2a_pb2.TaskArtifactUpdateEvent(
            task_id=event.task_id,
            context_id=event.context_id,
            artifact=cls.artifact(event.artifact),
            metadata=cls.metadata(event.metadata),
            append=event.append or False,
            last_chunk=event.last_chunk or False,
        )

    @classmethod
    def task_status_update_event(
        cls, event: types.TaskStatusUpdateEvent
    ) -> a2a_pb2.TaskStatusUpdateEvent:
        return a2a_pb2.TaskStatusUpdateEvent(
            task_id=event.task_id,
            context_id=event.context_id,
            status=cls.task_status(event.status),
            metadata=cls.metadata(event.metadata),
            final=event.final,
        )

    @classmethod
    def message_send_configuration(
        cls, config: types.MessageSendConfiguration | None
    ) -> a2a_pb2.SendMessageConfiguration:
        if not config:
            return a2a_pb2.SendMessageConfiguration()
        return a2a_pb2.SendMessageConfiguration(
            accepted_output_modes=config.accepted_output_modes,
            push_notification=cls.push_notification_config(
                config.push_notification_config
            )
            if config.push_notification_config
            else None,
            history_length=config.history_length,
            blocking=config.blocking or False,
        )

    @classmethod
    def update_event(
        cls,
        event: types.Task
        | types.Message
        | types.TaskStatusUpdateEvent
        | types.TaskArtifactUpdateEvent,
    ) -> a2a_pb2.StreamResponse:
        """Converts a task, message, or task update event to a StreamResponse."""
        return cls.stream_response(event)

    @classmethod
    def task_or_message(
        cls, event: types.Task | types.Message
    ) -> a2a_pb2.SendMessageResponse:
        if isinstance(event, types.Message):
            return a2a_pb2.SendMessageResponse(
                msg=cls.message(event),
            )
        return a2a_pb2.SendMessageResponse(
            task=cls.task(event),
        )

    @classmethod
    def stream_response(
        cls,
        event: (
            types.Message
            | types.Task
            | types.TaskStatusUpdateEvent
            | types.TaskArtifactUpdateEvent
        ),
    ) -> a2a_pb2.StreamResponse:
        if isinstance(event, types.Message):
            return a2a_pb2.StreamResponse(msg=cls.message(event))
        if isinstance(event, types.Task):
            return a2a_pb2.StreamResponse(task=cls.task(event))
        if isinstance(event, types.TaskStatusUpdateEvent):
            return a2a_pb2.StreamResponse(
                status_update=cls.task_status_update_event(event),
            )
        if isinstance(event, types.TaskArtifactUpdateEvent):
            return a2a_pb2.StreamResponse(
                artifact_update=cls.task_artifact_update_event(event),
            )
        raise ValueError(f'Unsupported event type: {type(event)}')

    @classmethod
    def task_push_notification_config(
        cls, config: types.TaskPushNotificationConfig
    ) -> a2a_pb2.TaskPushNotificationConfig:
        return a2a_pb2.TaskPushNotificationConfig(
            name=f'tasks/{config.task_id}/pushNotificationConfigs/{config.push_notification_config.id}',
            push_notification_config=cls.push_notification_config(
                config.push_notification_config,
            ),
        )

    @classmethod
    def agent_card(
        cls,
        card: types.AgentCard,
    ) -> a2a_pb2.AgentCard:
        return a2a_pb2.AgentCard(
            capabilities=cls.capabilities(card.capabilities),
            default_input_modes=list(card.default_input_modes),
            default_output_modes=list(card.default_output_modes),
            description=card.description,
            documentation_url=card.documentation_url,
            name=card.name,
            provider=cls.provider(card.provider),
            security=cls.security(card.security),
            security_schemes=cls.security_schemes(card.security_schemes),
            skills=[cls.skill(x) for x in card.skills] if card.skills else [],
            url=card.url,
            version=card.version,
            supports_authenticated_extended_card=bool(
                card.supports_authenticated_extended_card
            ),
            preferred_transport=card.preferred_transport,
            protocol_version=card.protocol_version,
            additional_interfaces=[
                cls.agent_interface(x) for x in card.additional_interfaces
            ]
            if card.additional_interfaces
            else None,
            signatures=[cls.agent_card_signature(x) for x in card.signatures]
            if card.signatures
            else None,
        )

    @classmethod
    def agent_card_signature(
        cls, signature: types.AgentCardSignature
    ) -> a2a_pb2.AgentCardSignature:
        return a2a_pb2.AgentCardSignature(
            protected=signature.protected,
            signature=signature.signature,
            header=dict_to_struct(signature.header)
            if signature.header is not None
            else None,
        )

    @classmethod
    def agent_interface(
        cls,
        interface: types.AgentInterface,
    ) -> a2a_pb2.AgentInterface:
        return a2a_pb2.AgentInterface(
            transport=interface.transport,
            url=interface.url,
        )

    @classmethod
    def capabilities(
        cls, capabilities: types.AgentCapabilities
    ) -> a2a_pb2.AgentCapabilities:
        return a2a_pb2.AgentCapabilities(
            streaming=bool(capabilities.streaming),
            push_notifications=bool(capabilities.push_notifications),
            extensions=[
                cls.extension(x) for x in capabilities.extensions or []
            ],
        )

    @classmethod
    def extension(
        cls,
        extension: types.AgentExtension,
    ) -> a2a_pb2.AgentExtension:
        return a2a_pb2.AgentExtension(
            uri=extension.uri,
            description=extension.description,
            params=dict_to_struct(extension.params)
            if extension.params
            else None,
            required=extension.required,
        )

    @classmethod
    def provider(
        cls, provider: types.AgentProvider | None
    ) -> a2a_pb2.AgentProvider | None:
        if not provider:
            return None
        return a2a_pb2.AgentProvider(
            organization=provider.organization,
            url=provider.url,
        )

    @classmethod
    def security(
        cls,
        security: list[dict[str, list[str]]] | None,
    ) -> list[a2a_pb2.Security] | None:
        if not security:
            return None
        return [
            a2a_pb2.Security(
                schemes={k: a2a_pb2.StringList(list=v) for (k, v) in s.items()}
            )
            for s in security
        ]

    @classmethod
    def security_schemes(
        cls,
        schemes: dict[str, types.SecurityScheme] | None,
    ) -> dict[str, a2a_pb2.SecurityScheme] | None:
        if not schemes:
            return None
        return {k: cls.security_scheme(v) for (k, v) in schemes.items()}

    @classmethod
    def security_scheme(
        cls,
        scheme: types.SecurityScheme,
    ) -> a2a_pb2.SecurityScheme:
        if isinstance(scheme.root, types.APIKeySecurityScheme):
            return a2a_pb2.SecurityScheme(
                api_key_security_scheme=a2a_pb2.APIKeySecurityScheme(
                    description=scheme.root.description,
                    location=scheme.root.in_.value,
                    name=scheme.root.name,
                )
            )
        if isinstance(scheme.root, types.HTTPAuthSecurityScheme):
            return a2a_pb2.SecurityScheme(
                http_auth_security_scheme=a2a_pb2.HTTPAuthSecurityScheme(
                    description=scheme.root.description,
                    scheme=scheme.root.scheme,
                    bearer_format=scheme.root.bearer_format,
                )
            )
        if isinstance(scheme.root, types.OAuth2SecurityScheme):
            return a2a_pb2.SecurityScheme(
                oauth2_security_scheme=a2a_pb2.OAuth2SecurityScheme(
                    description=scheme.root.description,
                    flows=cls.oauth2_flows(scheme.root.flows),
                )
            )
        if isinstance(scheme.root, types.MutualTLSSecurityScheme):
            return a2a_pb2.SecurityScheme(
                mtls_security_scheme=a2a_pb2.MutualTlsSecurityScheme(
                    description=scheme.root.description,
                )
            )
        return a2a_pb2.SecurityScheme(
            open_id_connect_security_scheme=a2a_pb2.OpenIdConnectSecurityScheme(
                description=scheme.root.description,
                open_id_connect_url=scheme.root.open_id_connect_url,
            )
        )

    @classmethod
    def oauth2_flows(cls, flows: types.OAuthFlows) -> a2a_pb2.OAuthFlows:
        if flows.authorization_code:
            return a2a_pb2.OAuthFlows(
                authorization_code=a2a_pb2.AuthorizationCodeOAuthFlow(
                    authorization_url=flows.authorization_code.authorization_url,
                    refresh_url=flows.authorization_code.refresh_url,
                    scopes=dict(flows.authorization_code.scopes.items()),
                    token_url=flows.authorization_code.token_url,
                ),
            )
        if flows.client_credentials:
            return a2a_pb2.OAuthFlows(
                client_credentials=a2a_pb2.ClientCredentialsOAuthFlow(
                    refresh_url=flows.client_credentials.refresh_url,
                    scopes=dict(flows.client_credentials.scopes.items()),
                    token_url=flows.client_credentials.token_url,
                ),
            )
        if flows.implicit:
            return a2a_pb2.OAuthFlows(
                implicit=a2a_pb2.ImplicitOAuthFlow(
                    authorization_url=flows.implicit.authorization_url,
                    refresh_url=flows.implicit.refresh_url,
                    scopes=dict(flows.implicit.scopes.items()),
                ),
            )
        if flows.password:
            return a2a_pb2.OAuthFlows(
                password=a2a_pb2.PasswordOAuthFlow(
                    refresh_url=flows.password.refresh_url,
                    scopes=dict(flows.password.scopes.items()),
                    token_url=flows.password.token_url,
                ),
            )
        raise ValueError('Unknown oauth flow definition')

    @classmethod
    def skill(cls, skill: types.AgentSkill) -> a2a_pb2.AgentSkill:
        return a2a_pb2.AgentSkill(
            id=skill.id,
            name=skill.name,
            description=skill.description,
            tags=skill.tags,
            examples=skill.examples,
            input_modes=skill.input_modes,
            output_modes=skill.output_modes,
        )

    @classmethod
    def role(cls, role: types.Role) -> a2a_pb2.Role:
        match role:
            case types.Role.user:
                return a2a_pb2.Role.ROLE_USER
            case types.Role.agent:
                return a2a_pb2.Role.ROLE_AGENT
            case _:
                return a2a_pb2.Role.ROLE_UNSPECIFIED


class FromProto:
    """Converts proto types to Python types."""

    @classmethod
    def message(cls, message: a2a_pb2.Message) -> types.Message:
        return types.Message(
            message_id=message.message_id,
            parts=[cls.part(p) for p in message.content],
            context_id=message.context_id or None,
            task_id=message.task_id or None,
            role=cls.role(message.role),
            metadata=cls.metadata(message.metadata),
            extensions=list(message.extensions) or None,
        )

    @classmethod
    def metadata(cls, metadata: struct_pb2.Struct) -> dict[str, Any]:
        if not metadata.fields:
            return {}
        return json_format.MessageToDict(metadata)

    @classmethod
    def part(cls, part: a2a_pb2.Part) -> types.Part:
        if part.HasField('text'):
            return types.Part(
                root=types.TextPart(
                    text=part.text,
                    metadata=cls.metadata(part.metadata)
                    if part.metadata
                    else None,
                ),
            )
        if part.HasField('file'):
            return types.Part(
                root=types.FilePart(
                    file=cls.file(part.file),
                    metadata=cls.metadata(part.metadata)
                    if part.metadata
                    else None,
                ),
            )
        if part.HasField('data'):
            return types.Part(
                root=types.DataPart(
                    data=cls.data(part.data),
                    metadata=cls.metadata(part.metadata)
                    if part.metadata
                    else None,
                ),
            )
        raise ValueError(f'Unsupported part type: {part}')

    @classmethod
    def data(cls, data: a2a_pb2.DataPart) -> dict[str, Any]:
        json_data = json_format.MessageToJson(data.data)
        return json.loads(json_data)

    @classmethod
    def file(
        cls, file: a2a_pb2.FilePart
    ) -> types.FileWithUri | types.FileWithBytes:
        common_args = {
            'mime_type': file.mime_type or None,
            'name': file.name or None,
        }
        if file.HasField('file_with_uri'):
            return types.FileWithUri(
                uri=file.file_with_uri,
                **common_args,
            )
        return types.FileWithBytes(
            bytes=file.file_with_bytes.decode('utf-8'),
            **common_args,
        )

    @classmethod
    def task_or_message(
        cls, event: a2a_pb2.SendMessageResponse
    ) -> types.Task | types.Message:
        if event.HasField('msg'):
            return cls.message(event.msg)
        return cls.task(event.task)

    @classmethod
    def task(cls, task: a2a_pb2.Task) -> types.Task:
        return types.Task(
            id=task.id,
            context_id=task.context_id,
            status=cls.task_status(task.status),
            artifacts=[cls.artifact(a) for a in task.artifacts],
            history=[cls.message(h) for h in task.history],
            metadata=cls.metadata(task.metadata),
        )

    @classmethod
    def task_status(cls, status: a2a_pb2.TaskStatus) -> types.TaskStatus:
        return types.TaskStatus(
            state=cls.task_state(status.state),
            message=cls.message(status.update),
        )

    @classmethod
    def task_state(cls, state: a2a_pb2.TaskState) -> types.TaskState:
        match state:
            case a2a_pb2.TaskState.TASK_STATE_SUBMITTED:
                return types.TaskState.submitted
            case a2a_pb2.TaskState.TASK_STATE_WORKING:
                return types.TaskState.working
            case a2a_pb2.TaskState.TASK_STATE_COMPLETED:
                return types.TaskState.completed
            case a2a_pb2.TaskState.TASK_STATE_CANCELLED:
                return types.TaskState.canceled
            case a2a_pb2.TaskState.TASK_STATE_FAILED:
                return types.TaskState.failed
            case a2a_pb2.TaskState.TASK_STATE_INPUT_REQUIRED:
                return types.TaskState.input_required
            case a2a_pb2.TaskState.TASK_STATE_AUTH_REQUIRED:
                return types.TaskState.auth_required
            case a2a_pb2.TaskState.TASK_STATE_REJECTED:
                return types.TaskState.rejected
            case _:
                return types.TaskState.unknown

    @classmethod
    def artifact(cls, artifact: a2a_pb2.Artifact) -> types.Artifact:
        return types.Artifact(
            artifact_id=artifact.artifact_id,
            description=artifact.description,
            metadata=cls.metadata(artifact.metadata),
            name=artifact.name,
            parts=[cls.part(p) for p in artifact.parts],
            extensions=artifact.extensions or None,
        )

    @classmethod
    def task_artifact_update_event(
        cls, event: a2a_pb2.TaskArtifactUpdateEvent
    ) -> types.TaskArtifactUpdateEvent:
        return types.TaskArtifactUpdateEvent(
            task_id=event.task_id,
            context_id=event.context_id,
            artifact=cls.artifact(event.artifact),
            metadata=cls.metadata(event.metadata),
            append=event.append,
            last_chunk=event.last_chunk,
        )

    @classmethod
    def task_status_update_event(
        cls, event: a2a_pb2.TaskStatusUpdateEvent
    ) -> types.TaskStatusUpdateEvent:
        return types.TaskStatusUpdateEvent(
            task_id=event.task_id,
            context_id=event.context_id,
            status=cls.task_status(event.status),
            metadata=cls.metadata(event.metadata),
            final=event.final,
        )

    @classmethod
    def push_notification_config(
        cls, config: a2a_pb2.PushNotificationConfig
    ) -> types.PushNotificationConfig:
        return types.PushNotificationConfig(
            id=config.id,
            url=config.url,
            token=config.token,
            authentication=cls.authentication_info(config.authentication)
            if config.HasField('authentication')
            else None,
        )

    @classmethod
    def authentication_info(
        cls, info: a2a_pb2.AuthenticationInfo
    ) -> types.PushNotificationAuthenticationInfo:
        return types.PushNotificationAuthenticationInfo(
            schemes=list(info.schemes),
            credentials=info.credentials,
        )

    @classmethod
    def message_send_configuration(
        cls, config: a2a_pb2.SendMessageConfiguration
    ) -> types.MessageSendConfiguration:
        return types.MessageSendConfiguration(
            accepted_output_modes=list(config.accepted_output_modes),
            push_notification_config=cls.push_notification_config(
                config.push_notification
            )
            if config.HasField('push_notification')
            else None,
            history_length=config.history_length,
            blocking=config.blocking,
        )

    @classmethod
    def message_send_params(
        cls, request: a2a_pb2.SendMessageRequest
    ) -> types.MessageSendParams:
        return types.MessageSendParams(
            configuration=cls.message_send_configuration(request.configuration),
            message=cls.message(request.request),
            metadata=cls.metadata(request.metadata),
        )

    @classmethod
    def task_id_params(
        cls,
        request: (
            a2a_pb2.CancelTaskRequest
            | a2a_pb2.TaskSubscriptionRequest
            | a2a_pb2.GetTaskPushNotificationConfigRequest
        ),
    ) -> types.TaskIdParams:
        if isinstance(request, a2a_pb2.GetTaskPushNotificationConfigRequest):
            m = TASK_PUSH_CONFIG_NAME_MATCH.match(request.name)
            if not m:
                raise InvalidParamsError(message=f'No task for {request.name}')
            return types.TaskIdParams(id=m.group(1))
        m = TASK_NAME_MATCH.match(request.name)
        if not m:
            raise InvalidParamsError(message=f'No task for {request.name}')
        return types.TaskIdParams(id=m.group(1))

    @classmethod
    def task_push_notification_config_request(
        cls,
        request: a2a_pb2.CreateTaskPushNotificationConfigRequest,
    ) -> types.TaskPushNotificationConfig:
        m = TASK_NAME_MATCH.match(request.parent)
        if not m:
            raise InvalidParamsError(message=f'No task for {request.parent}')
        return types.TaskPushNotificationConfig(
            push_notification_config=cls.push_notification_config(
                request.config.push_notification_config,
            ),
            task_id=m.group(1),
        )

    @classmethod
    def task_push_notification_config(
        cls,
        config: a2a_pb2.TaskPushNotificationConfig,
    ) -> types.TaskPushNotificationConfig:
        m = TASK_PUSH_CONFIG_NAME_MATCH.match(config.name)
        if not m:
            raise InvalidParamsError(
                message=f'Bad TaskPushNotificationConfig resource name {config.name}'
            )
        return types.TaskPushNotificationConfig(
            push_notification_config=cls.push_notification_config(
                config.push_notification_config,
            ),
            task_id=m.group(1),
        )

    @classmethod
    def agent_card(
        cls,
        card: a2a_pb2.AgentCard,
    ) -> types.AgentCard:
        return types.AgentCard(
            capabilities=cls.capabilities(card.capabilities),
            default_input_modes=list(card.default_input_modes),
            default_output_modes=list(card.default_output_modes),
            description=card.description,
            documentation_url=card.documentation_url,
            name=card.name,
            provider=cls.provider(card.provider),
            security=cls.security(list(card.security)),
            security_schemes=cls.security_schemes(dict(card.security_schemes)),
            skills=[cls.skill(x) for x in card.skills] if card.skills else [],
            url=card.url,
            version=card.version,
            supports_authenticated_extended_card=card.supports_authenticated_extended_card,
            preferred_transport=card.preferred_transport,
            protocol_version=card.protocol_version,
            additional_interfaces=[
                cls.agent_interface(x) for x in card.additional_interfaces
            ]
            if card.additional_interfaces
            else None,
            signatures=[cls.agent_card_signature(x) for x in card.signatures]
            if card.signatures
            else None,
        )

    @classmethod
    def agent_card_signature(
        cls, signature: a2a_pb2.AgentCardSignature
    ) -> types.AgentCardSignature:
        return types.AgentCardSignature(
            protected=signature.protected,
            signature=signature.signature,
            header=json_format.MessageToDict(signature.header),
        )

    @classmethod
    def agent_interface(
        cls,
        interface: a2a_pb2.AgentInterface,
    ) -> types.AgentInterface:
        return types.AgentInterface(
            transport=interface.transport,
            url=interface.url,
        )

    @classmethod
    def task_query_params(
        cls,
        request: a2a_pb2.GetTaskRequest,
    ) -> types.TaskQueryParams:
        m = TASK_NAME_MATCH.match(request.name)
        if not m:
            raise InvalidParamsError(message=f'No task for {request.name}')
        return types.TaskQueryParams(
            history_length=request.history_length
            if request.history_length
            else None,
            id=m.group(1),
            metadata=None,
        )

    @classmethod
    def capabilities(
        cls, capabilities: a2a_pb2.AgentCapabilities
    ) -> types.AgentCapabilities:
        return types.AgentCapabilities(
            streaming=capabilities.streaming,
            push_notifications=capabilities.push_notifications,
            extensions=[
                cls.agent_extension(x) for x in capabilities.extensions
            ],
        )

    @classmethod
    def agent_extension(
        cls,
        extension: a2a_pb2.AgentExtension,
    ) -> types.AgentExtension:
        return types.AgentExtension(
            uri=extension.uri,
            description=extension.description,
            params=json_format.MessageToDict(extension.params),
            required=extension.required,
        )

    @classmethod
    def security(
        cls,
        security: list[a2a_pb2.Security] | None,
    ) -> list[dict[str, list[str]]] | None:
        if not security:
            return None
        return [
            {k: list(v.list) for (k, v) in s.schemes.items()} for s in security
        ]

    @classmethod
    def provider(
        cls, provider: a2a_pb2.AgentProvider | None
    ) -> types.AgentProvider | None:
        if not provider:
            return None
        return types.AgentProvider(
            organization=provider.organization,
            url=provider.url,
        )

    @classmethod
    def security_schemes(
        cls, schemes: dict[str, a2a_pb2.SecurityScheme]
    ) -> dict[str, types.SecurityScheme]:
        return {k: cls.security_scheme(v) for (k, v) in schemes.items()}

    @classmethod
    def security_scheme(
        cls,
        scheme: a2a_pb2.SecurityScheme,
    ) -> types.SecurityScheme:
        if scheme.HasField('api_key_security_scheme'):
            return types.SecurityScheme(
                root=types.APIKeySecurityScheme(
                    description=scheme.api_key_security_scheme.description,
                    name=scheme.api_key_security_scheme.name,
                    in_=types.In(scheme.api_key_security_scheme.location),  # type: ignore[call-arg]
                )
            )
        if scheme.HasField('http_auth_security_scheme'):
            return types.SecurityScheme(
                root=types.HTTPAuthSecurityScheme(
                    description=scheme.http_auth_security_scheme.description,
                    scheme=scheme.http_auth_security_scheme.scheme,
                    bearer_format=scheme.http_auth_security_scheme.bearer_format,
                )
            )
        if scheme.HasField('oauth2_security_scheme'):
            return types.SecurityScheme(
                root=types.OAuth2SecurityScheme(
                    description=scheme.oauth2_security_scheme.description,
                    flows=cls.oauth2_flows(scheme.oauth2_security_scheme.flows),
                )
            )
        if scheme.HasField('mtls_security_scheme'):
            return types.SecurityScheme(
                root=types.MutualTLSSecurityScheme(
                    description=scheme.mtls_security_scheme.description,
                )
            )
        return types.SecurityScheme(
            root=types.OpenIdConnectSecurityScheme(
                description=scheme.open_id_connect_security_scheme.description,
                open_id_connect_url=scheme.open_id_connect_security_scheme.open_id_connect_url,
            )
        )

    @classmethod
    def oauth2_flows(cls, flows: a2a_pb2.OAuthFlows) -> types.OAuthFlows:
        if flows.HasField('authorization_code'):
            return types.OAuthFlows(
                authorization_code=types.AuthorizationCodeOAuthFlow(
                    authorization_url=flows.authorization_code.authorization_url,
                    refresh_url=flows.authorization_code.refresh_url,
                    scopes=dict(flows.authorization_code.scopes.items()),
                    token_url=flows.authorization_code.token_url,
                ),
            )
        if flows.HasField('client_credentials'):
            return types.OAuthFlows(
                client_credentials=types.ClientCredentialsOAuthFlow(
                    refresh_url=flows.client_credentials.refresh_url,
                    scopes=dict(flows.client_credentials.scopes.items()),
                    token_url=flows.client_credentials.token_url,
                ),
            )
        if flows.HasField('implicit'):
            return types.OAuthFlows(
                implicit=types.ImplicitOAuthFlow(
                    authorization_url=flows.implicit.authorization_url,
                    refresh_url=flows.implicit.refresh_url,
                    scopes=dict(flows.implicit.scopes.items()),
                ),
            )
        return types.OAuthFlows(
            password=types.PasswordOAuthFlow(
                refresh_url=flows.password.refresh_url,
                scopes=dict(flows.password.scopes.items()),
                token_url=flows.password.token_url,
            ),
        )

    @classmethod
    def stream_response(
        cls,
        response: a2a_pb2.StreamResponse,
    ) -> (
        types.Message
        | types.Task
        | types.TaskStatusUpdateEvent
        | types.TaskArtifactUpdateEvent
    ):
        if response.HasField('msg'):
            return cls.message(response.msg)
        if response.HasField('task'):
            return cls.task(response.task)
        if response.HasField('status_update'):
            return cls.task_status_update_event(response.status_update)
        if response.HasField('artifact_update'):
            return cls.task_artifact_update_event(response.artifact_update)
        raise ValueError('Unsupported StreamResponse type')

    @classmethod
    def list_task_push_notification_config_response(
        cls, response: a2a_pb2.ListTaskPushNotificationConfigResponse
    ) -> types.ListTaskPushNotificationConfigResponse:
        return types.ListTaskPushNotificationConfigResponse(
            root=types.ListTaskPushNotificationConfigSuccessResponse(
                result=[
                    cls.task_push_notification_config(c)
                    for c in response.configs
                ],
                id=None,
            )
        )

    @classmethod
    def skill(cls, skill: a2a_pb2.AgentSkill) -> types.AgentSkill:
        return types.AgentSkill(
            id=skill.id,
            name=skill.name,
            description=skill.description,
            tags=list(skill.tags),
            examples=list(skill.examples),
            input_modes=list(skill.input_modes),
            output_modes=list(skill.output_modes),
        )

    @classmethod
    def role(cls, role: a2a_pb2.Role) -> types.Role:
        match role:
            case a2a_pb2.Role.ROLE_USER:
                return types.Role.user
            case a2a_pb2.Role.ROLE_AGENT:
                return types.Role.agent
            case _:
                return types.Role.agent



from __future__ import annotations

import asyncio
import logging

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from a2a.server.agent_execution.agent_executor import AgentExecutor
    from a2a.server.context import ServerCallContext
    from a2a.server.tasks.push_notification_sender import PushNotificationSender
    from a2a.server.tasks.task_store import TaskStore

from a2a.server.agent_execution.active_task import ActiveTask
from a2a.server.tasks.task_manager import TaskManager


logger = logging.getLogger(__name__)


class ActiveTaskRegistry:
    """A registry for active ActiveTask instances."""

    def __init__(
        self,
        agent_executor: AgentExecutor,
        task_store: TaskStore,
        push_sender: PushNotificationSender | None = None,
    ):
        self._agent_executor = agent_executor
        self._task_store = task_store
        self._push_sender = push_sender
        self._active_tasks: dict[str, ActiveTask] = {}
        self._lock = asyncio.Lock()
        self._cleanup_tasks: set[asyncio.Task[None]] = set()

    async def get_or_create(
        self,
        task_id: str,
        call_context: ServerCallContext,
        context_id: str | None = None,
        create_task_if_missing: bool = False,
    ) -> ActiveTask:
        """Retrieves an existing ActiveTask or creates a new one."""
        async with self._lock:
            if task_id in self._active_tasks:
                return self._active_tasks[task_id]

            task_manager = TaskManager(
                task_id=task_id,
                context_id=context_id,
                task_store=self._task_store,
                initial_message=None,
                context=call_context,
            )

            active_task = ActiveTask(
                agent_executor=self._agent_executor,
                task_id=task_id,
                task_manager=task_manager,
                push_sender=self._push_sender,
                on_cleanup=self._on_active_task_cleanup,
            )
            self._active_tasks[task_id] = active_task

        await active_task.start(
            call_context=call_context,
            create_task_if_missing=create_task_if_missing,
        )
        return active_task

    def _on_active_task_cleanup(self, active_task: ActiveTask) -> None:
        """Called by ActiveTask when it's finished and has no subscribers."""
        logger.debug('Active task %s cleanup scheduled', active_task.task_id)
        task = asyncio.create_task(self._remove_task(active_task.task_id))
        self._cleanup_tasks.add(task)
        task.add_done_callback(self._cleanup_tasks.discard)

    async def _remove_task(self, task_id: str) -> None:
        async with self._lock:
            self._active_tasks.pop(task_id, None)
            logger.debug('Removed active task for %s from registry', task_id)

    async def get(self, task_id: str) -> ActiveTask | None:
        """Retrieves an existing task."""
        async with self._lock:
            return self._active_tasks.get(task_id)



# ruff: noqa: TRY301, SLF001
from __future__ import annotations

import asyncio
import logging
import uuid

from typing import TYPE_CHECKING, Any, cast

from a2a.server.agent_execution.context import RequestContext


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable

    from a2a.server.agent_execution.agent_executor import AgentExecutor
    from a2a.server.context import ServerCallContext
    from a2a.server.tasks.push_notification_sender import (
        PushNotificationSender,
    )
    from a2a.server.tasks.task_manager import TaskManager

from a2a.server.events.event_queue_v2 import (
    AsyncQueue,
    Event,
    EventQueueSource,
    QueueShutDown,
    _create_async_queue,
)
from a2a.server.tasks import PushNotificationEvent
from a2a.types.a2a_pb2 import (
    Message,
    Task,
    TaskState,
    TaskStatus,
    TaskStatusUpdateEvent,
)
from a2a.utils.errors import (
    InvalidAgentResponseError,
    InvalidParamsError,
    TaskNotFoundError,
)


logger = logging.getLogger(__name__)


TERMINAL_TASK_STATES = {
    TaskState.TASK_STATE_COMPLETED,
    TaskState.TASK_STATE_CANCELED,
    TaskState.TASK_STATE_FAILED,
    TaskState.TASK_STATE_REJECTED,
}
INTERRUPTED_TASK_STATES = {
    TaskState.TASK_STATE_AUTH_REQUIRED,
    TaskState.TASK_STATE_INPUT_REQUIRED,
}


class _RequestStarted:
    def __init__(self, request_id: uuid.UUID, request_context: RequestContext):
        self.request_id = request_id
        self.request_context = request_context


class _RequestCompleted:
    def __init__(self, request_id: uuid.UUID):
        self.request_id = request_id


class ActiveTask:
    """Manages the lifecycle and execution of an active A2A task.

    It coordinates between the agent's execution (the producer), the
    persistence and state management (the TaskManager), and the event
    distribution to subscribers (the consumer).

    Concurrency Guarantees:
    - This class is designed to be highly concurrent. It manages an internal
      producer-consumer model using `asyncio.Task`s.
    - `self._lock` (asyncio.Lock) ensures mutually exclusive access for critical
      lifecycle state changes, such as starting the task, subscribing, and
      determining if cleanup is safe to trigger.

      mutation to the observable result state (like `_exception`,
      or `_is_finished`) notifies waiting coroutines (like `wait()`).
    - `self._is_finished` (asyncio.Event) provides a thread-safe, non-blocking way
      for external observers and internal loops to check if the ActiveTask has
      permanently ceased execution and closed its queues.
    """

    def __init__(
        self,
        agent_executor: AgentExecutor,
        task_id: str,
        task_manager: TaskManager,
        push_sender: PushNotificationSender | None = None,
        on_cleanup: Callable[[ActiveTask], None] | None = None,
    ) -> None:
        """Initializes the ActiveTask.

        Args:
            agent_executor: The executor to run the agent logic (producer).
            task_id: The unique identifier of the task being managed.
            task_manager: The manager for task state and database persistence.
            push_sender: Optional sender for out-of-band push notifications.
            on_cleanup: Optional callback triggered when the task is fully finished
                        and the last subscriber has disconnected. Used to prune
                        the task from the ActiveTaskRegistry.
        """
        # --- Core Dependencies ---
        self._agent_executor = agent_executor
        self._task_id = task_id
        self._event_queue_agent = EventQueueSource()
        self._event_queue_subscribers = EventQueueSource(
            create_default_sink=False
        )
        self._task_manager = task_manager
        self._push_sender = push_sender
        self._on_cleanup = on_cleanup

        # --- Synchronization Primitives ---
        # `_lock` protects structural lifecycle changes: start(), subscribe() counting,
        # and _maybe_cleanup() race conditions.
        self._lock = asyncio.Lock()

        # `_request_lock` protects parallel request processing.
        self._request_lock = asyncio.Lock()

        # _task_created is set when initial version of task is stored in DB.
        self._task_created = asyncio.Event()

        # `_is_finished` is set EXACTLY ONCE when the consumer loop exits, signifying
        # the absolute end of the task's active lifecycle.
        self._is_finished = asyncio.Event()

        # --- Lifecycle State ---
        # The background task executing the agent logic.
        self._producer_task: asyncio.Task[None] | None = None
        # The background task reading from _event_queue and updating the DB.
        self._consumer_task: asyncio.Task[None] | None = None

        # Tracks how many active SSE/gRPC streams are currently tailing this task.
        # Protected by `_lock`.
        self._reference_count = 0

        # Holds any fatal exception that crashed the producer or consumer.
        # TODO: Synchronize exception handling (ideally mix it in the queue).
        self._exception: Exception | None = None

        # Queue for incoming requests
        self._request_queue: AsyncQueue[tuple[RequestContext, uuid.UUID]] = (
            _create_async_queue()
        )

    @property
    def task_id(self) -> str:
        """The ID of the task."""
        return self._task_id

    async def enqueue_request(
        self, request_context: RequestContext
    ) -> uuid.UUID:
        """Enqueues a request for the active task to process."""
        request_id = uuid.uuid4()
        await self._request_queue.put((request_context, request_id))
        return request_id

    async def start(
        self,
        call_context: ServerCallContext,
        create_task_if_missing: bool = False,
    ) -> None:
        """Starts the active task background processes.

        Concurrency Guarantee:
        Uses `self._lock` to ensure the producer and consumer tasks are strictly
        singleton instances for the lifetime of this ActiveTask.
        """
        logger.debug('ActiveTask[%s]: Starting', self._task_id)
        async with self._lock:
            if self._is_finished.is_set():
                raise InvalidParamsError(
                    f'Task {self._task_id} is already completed. Cannot start it again.'
                )

            if (
                self._producer_task is not None
                and self._consumer_task is not None
            ):
                logger.debug(
                    'ActiveTask[%s]: Already started, ignoring start request',
                    self._task_id,
                )
                return

            logger.debug(
                'ActiveTask[%s]: Executing setup (call_context: %s, create_task_if_missing: %s)',
                self._task_id,
                call_context,
                create_task_if_missing,
            )
            try:
                self._task_manager._call_context = call_context
                task = await self._task_manager.get_task()
                logger.debug('TASK (start): %s', task)

                if task:
                    self._task_created.set()
                    if task.status.state in TERMINAL_TASK_STATES:
                        raise InvalidParamsError(
                            message=f'Task {task.id} is in terminal state: {task.status.state}'
                        )
                elif not create_task_if_missing:
                    raise TaskNotFoundError

            except Exception:
                logger.debug(
                    'ActiveTask[%s]: Setup failed, cleaning up',
                    self._task_id,
                )
                self._is_finished.set()
                if self._reference_count == 0 and self._on_cleanup:
                    self._on_cleanup(self)
                raise

            # Spawn the background tasks that drive the lifecycle.
            self._reference_count += 1
            self._producer_task = asyncio.create_task(
                self._run_producer(), name=f'producer:{self._task_id}'
            )
            self._consumer_task = asyncio.create_task(
                self._run_consumer(), name=f'consumer:{self._task_id}'
            )
            logger.debug(
                'ActiveTask[%s]: Background tasks created', self._task_id
            )

    async def _run_producer(self) -> None:
        """Executes the agent logic.

        This method encapsulates the external `AgentExecutor.execute` call. It ensures
        that regardless of how the agent finishes (success, unhandled exception, or
        cancellation), the underlying `_event_queue` is safely closed, which signals
        the consumer to wind down.

        Concurrency Guarantee:
        Runs as a detached asyncio.Task. Safe to cancel.
        """
        logger.debug('Producer[%s]: Started', self._task_id)
        request_context = None
        try:
            while True:
                (
                    request_context,
                    request_id,
                ) = await self._request_queue.get()
                await self._request_lock.acquire()
                # TODO: Should we create task manager every time?
                self._task_manager._call_context = request_context.call_context

                request_context.current_task = (
                    await self._task_manager.get_task()
                )

                logger.debug(
                    'Producer[%s]: Executing agent task %s',
                    self._task_id,
                    request_context.current_task,
                )

                try:
                    await self._event_queue_agent.enqueue_event(
                        cast(
                            'Event',
                            _RequestStarted(request_id, request_context),
                        )
                    )

                    await self._agent_executor.execute(
                        request_context, self._event_queue_agent
                    )
                    logger.debug(
                        'Producer[%s]: Execution finished successfully',
                        self._task_id,
                    )
                finally:
                    logger.debug(
                        'Producer[%s]: Enqueuing request completed event',
                        self._task_id,
                    )
                    await self._event_queue_agent.enqueue_event(
                        cast('Event', _RequestCompleted(request_id))
                    )
                    self._request_queue.task_done()
        except asyncio.CancelledError:
            logger.debug('Producer[%s]: Cancelled', self._task_id)

        except QueueShutDown:
            logger.debug('Producer[%s]: Queue shut down', self._task_id)

        except Exception as e:
            logger.exception(
                'Producer[%s]: Execution failed',
                self._task_id,
            )
            # Create task and mark as failed.
            if request_context:
                await self._task_manager.ensure_task_id(
                    self._task_id,
                    request_context.context_id or '',
                )
                self._task_created.set()
            async with self._lock:
                await self._mark_task_as_failed(e)

        finally:
            self._request_queue.shutdown(immediate=True)
            await self._event_queue_agent.close(immediate=False)
            await self._event_queue_subscribers.close(immediate=False)
            logger.debug('Producer[%s]: Completed', self._task_id)

    async def _run_consumer(self) -> None:  # noqa: PLR0915, PLR0912
        """Consumes events from the agent and updates system state.

        This continuous loop dequeues events emitted by the producer, updates the
        database via `TaskManager`, and intercepts critical task states (e.g.,
        INPUT_REQUIRED, COMPLETED, FAILED) to cache the final result.

        Concurrency Guarantee:
        Runs as a detached asyncio.Task. The loop ends gracefully when the producer
        closes the queue (raising `QueueShutDown`). Upon termination, it formally sets
        `_is_finished`, unblocking all global subscribers and wait() calls.
        """
        logger.debug('Consumer[%s]: Started', self._task_id)
        task_mode = None
        message_to_save = None
        # TODO: Make helper methods
        # TODO: Support Task enqueue
        try:
            try:
                try:
                    while True:
                        # Dequeue event. This raises QueueShutDown when finished.
                        logger.debug(
                            'Consumer[%s]: Waiting for event',
                            self._task_id,
                        )
                        new_task = None
                        event = await self._event_queue_agent.dequeue_event()
                        logger.debug(
                            'Consumer[%s]: Dequeued event %s',
                            self._task_id,
                            type(event).__name__,
                        )

                        try:
                            if isinstance(event, _RequestCompleted):
                                logger.debug(
                                    'Consumer[%s]: Request completed',
                                    self._task_id,
                                )
                                self._request_lock.release()
                            elif isinstance(event, _RequestStarted):
                                logger.debug(
                                    'Consumer[%s]: Request started',
                                    self._task_id,
                                )
                                message_to_save = event.request_context.message

                            elif isinstance(event, Message):
                                if task_mode is not None:
                                    if task_mode:
                                        raise InvalidAgentResponseError(
                                            'Received Message object in task mode. Use TaskStatusUpdateEvent or TaskArtifactUpdateEvent instead.'
                                        )
                                    raise InvalidAgentResponseError(
                                        'Multiple Message objects received.'
                                    )
                                task_mode = False
                                logger.debug(
                                    'Consumer[%s]: Setting result to Message: %s',
                                    self._task_id,
                                    event,
                                )
                            else:
                                if task_mode is False:
                                    raise InvalidAgentResponseError(
                                        f'Received {type(event).__name__} in message mode. Use Task with TaskStatusUpdateEvent and TaskArtifactUpdateEvent instead.'
                                    )

                                if isinstance(event, Task):
                                    existing_task = (
                                        await self._task_manager.get_task()
                                    )
                                    if existing_task:
                                        logger.error(
                                            'Task %s already exists. Ignoring task replacement.',
                                            self._task_id,
                                        )
                                    else:
                                        await (
                                            self._task_manager.save_task_event(
                                                event
                                            )
                                        )
                                    # Initial task should already contain the message.
                                    message_to_save = None
                                else:
                                    if (
                                        isinstance(event, TaskStatusUpdateEvent)
                                        and not self._task_created.is_set()
                                    ):
                                        task = (
                                            await self._task_manager.get_task()
                                        )
                                        if task is None:
                                            raise InvalidAgentResponseError(
                                                f'Agent should enqueue Task before {type(event).__name__} event'
                                            )

                                    new_task = (
                                        await self._task_manager.ensure_task_id(
                                            self._task_id,
                                            event.context_id,
                                        )
                                    )

                                    if message_to_save is not None:
                                        new_task = self._task_manager.update_with_message(
                                            message_to_save,
                                            new_task,
                                        )
                                        await (
                                            self._task_manager.save_task_event(
                                                new_task
                                            )
                                        )
                                        message_to_save = None

                                task_mode = True
                                # Save structural events (like TaskStatusUpdate) to DB.

                                self._task_manager.context_id = event.context_id
                                if not isinstance(event, Task):
                                    await self._task_manager.process(event)

                                # Check for AUTH_REQUIRED or INPUT_REQUIRED or TERMINAL states
                                new_task = await self._task_manager.get_task()
                                if new_task is None:
                                    raise RuntimeError(
                                        f'Task {self.task_id} not found'
                                    )
                                if isinstance(event, Task):
                                    event = new_task
                                is_interrupted = (
                                    new_task.status.state
                                    in INTERRUPTED_TASK_STATES
                                )
                                is_terminal = (
                                    new_task.status.state
                                    in TERMINAL_TASK_STATES
                                )

                                # If we hit a breakpoint or terminal state, lock in the result.
                                if is_interrupted or is_terminal:
                                    logger.debug(
                                        'Consumer[%s]: Setting first result as Task (state=%s)',
                                        self._task_id,
                                        new_task.status.state,
                                    )

                                if is_terminal:
                                    logger.debug(
                                        'Consumer[%s]: Reached terminal state %s',
                                        self._task_id,
                                        new_task.status.state,
                                    )
                                    if not self._is_finished.is_set():
                                        async with self._lock:
                                            # TODO: what about _reference_count when task is failing?
                                            self._reference_count -= 1
                                    # _maybe_cleanup() is called in finally block.

                                    # Terminate the ActiveTask globally.
                                    self._is_finished.set()
                                    self._request_queue.shutdown(immediate=True)

                                if is_interrupted:
                                    logger.debug(
                                        'Consumer[%s]: Interrupted with state %s',
                                        self._task_id,
                                        new_task.status.state,
                                    )

                                if (
                                    self._push_sender
                                    and self._task_id
                                    and isinstance(event, PushNotificationEvent)
                                ):
                                    logger.debug(
                                        'Consumer[%s]: Sending push notification',
                                        self._task_id,
                                    )
                                    await self._push_sender.send_notification(
                                        self._task_id, event
                                    )

                                self._task_created.set()

                        finally:
                            if new_task is not None:
                                new_task_copy = Task()
                                new_task_copy.CopyFrom(new_task)
                                new_task = new_task_copy
                            if isinstance(event, Task):
                                new_task_copy = Task()
                                new_task_copy.CopyFrom(event)
                                event = new_task_copy

                            logger.debug(
                                'Consumer[%s]: Enqueuing\nEvent: %s\nNew Task: %s\n',
                                self._task_id,
                                event,
                                new_task,
                            )
                            await self._event_queue_subscribers.enqueue_event(
                                cast('Any', (event, new_task))
                            )
                            self._event_queue_agent.task_done()
                except QueueShutDown:
                    logger.debug(
                        'Consumer[%s]: Event queue shut down', self._task_id
                    )
            except Exception as e:
                logger.exception('Consumer[%s]: Failed', self._task_id)
                # TODO: Make the task in database as failed.
                async with self._lock:
                    await self._mark_task_as_failed(e)
            finally:
                # The consumer is dead. The ActiveTask is permanently finished.
                self._is_finished.set()
                self._request_queue.shutdown(immediate=True)
                await self._event_queue_agent.close(immediate=True)

                logger.debug('Consumer[%s]: Finishing', self._task_id)
                await self._maybe_cleanup()
        finally:
            logger.debug('Consumer[%s]: Completed', self._task_id)

    async def subscribe(  # noqa: PLR0912, PLR0915
        self,
        *,
        request: RequestContext | None = None,
        include_initial_task: bool = False,
        replace_status_update_with_task: bool = False,
    ) -> AsyncGenerator[Event, None]:
        """Creates a queue tap and yields events as they are produced.

        Concurrency Guarantee:
        Uses `_lock` to safely increment and decrement `_reference_count`.
        Safely detaches its queue tap when the client disconnects or the task finishes,
        triggering `_maybe_cleanup()` to potentially garbage collect the ActiveTask.
        """
        logger.debug('Subscribe[%s]: New subscriber', self._task_id)

        async with self._lock:
            if self._exception:
                logger.debug(
                    'Subscribe[%s]: Failed, exception already set',
                    self._task_id,
                )
                raise self._exception
            if self._is_finished.is_set():
                raise InvalidParamsError(
                    f'Task {self._task_id} is already completed.'
                )
            self._reference_count += 1
            logger.debug(
                'Subscribe[%s]: Subscribers count: %d',
                self._task_id,
                self._reference_count,
            )

        tapped_queue = await self._event_queue_subscribers.tap()
        request_id = await self.enqueue_request(request) if request else None

        try:
            if include_initial_task:
                logger.debug(
                    'Subscribe[%s]: Including initial task',
                    self._task_id,
                )
                task = await self.get_task()
                yield task

            while True:
                try:
                    if self._exception:
                        raise self._exception

                    dequeued = await tapped_queue.dequeue_event()
                    event, updated_task = cast('Any', dequeued)
                    logger.debug(
                        'Subscriber[%s]\nDequeued event %s\nUpdated task %s\n',
                        self._task_id,
                        event,
                        updated_task,
                    )
                    if replace_status_update_with_task and isinstance(
                        event, TaskStatusUpdateEvent
                    ):
                        logger.debug(
                            'Subscriber[%s]: Replacing TaskStatusUpdateEvent with Task: %s',
                            self._task_id,
                            updated_task,
                        )
                        event = updated_task
                    if self._exception:
                        raise self._exception from None
                    if isinstance(event, _RequestCompleted):
                        if (
                            request_id is not None
                            and event.request_id == request_id
                        ):
                            logger.debug(
                                'Subscriber[%s]: Request completed',
                                self._task_id,
                            )
                            return
                        continue
                    elif isinstance(event, _RequestStarted):
                        logger.debug(
                            'Subscriber[%s]: Request started',
                            self._task_id,
                        )
                        continue
                    try:
                        yield event
                    finally:
                        tapped_queue.task_done()
                except (QueueShutDown, asyncio.CancelledError):
                    if self._exception:
                        raise self._exception from None
                    break
        finally:
            logger.debug('Subscribe[%s]: Unsubscribing', self._task_id)
            await tapped_queue.close(immediate=True)
            async with self._lock:
                self._reference_count -= 1
            # Evaluate if this was the last subscriber on a finished task.
            await self._maybe_cleanup()

    async def cancel(self, call_context: ServerCallContext) -> Task:
        """Cancels the running active task.

        Concurrency Guarantee:
        Uses `_lock` to ensure we don't attempt to cancel a producer that is
        already winding down or hasn't started. It fires the cancellation signal
        and blocks until the consumer processes the cancellation events.
        """
        logger.debug('Cancel[%s]: Cancelling task', self._task_id)

        # TODO: Conflicts with call_context on the pending request.
        self._task_manager._call_context = call_context

        task = await self._task_manager.get_task()
        request_context = RequestContext(
            call_context=call_context,
            task_id=self._task_id,
            context_id=task.context_id if task else None,
            task=task,
        )

        async with self._lock:
            if not self._is_finished.is_set() and self._producer_task:
                logger.debug(
                    'Cancel[%s]: Cancelling producer task', self._task_id
                )
                self._producer_task.cancel()
                try:
                    await self._agent_executor.cancel(
                        request_context, self._event_queue_agent
                    )
                except Exception as e:
                    logger.exception(
                        'Cancel[%s]: Agent cancel failed', self._task_id
                    )
                    await self._mark_task_as_failed(e)
                    raise
            else:
                logger.debug(
                    'Cancel[%s]: Task already finished [%s] or producer not started [%s], not cancelling',
                    self._task_id,
                    self._is_finished.is_set(),
                    self._producer_task,
                )

        await self._is_finished.wait()
        task = await self._task_manager.get_task()
        if not task:
            raise RuntimeError('Task should have been created')
        return task

    async def _maybe_cleanup(self) -> None:
        """Triggers cleanup if task is finished and has no subscribers.

        Concurrency Guarantee:
        Protected by `_lock` to prevent race conditions where a new subscriber
        attaches at the exact moment the task decides to garbage collect itself.
        """
        async with self._lock:
            logger.debug(
                'Cleanup[%s]: Subscribers count: %d is_finished: %s',
                self._task_id,
                self._reference_count,
                self._is_finished.is_set(),
            )

            if (
                self._is_finished.is_set()
                and self._reference_count == 0
                and self._on_cleanup
            ):
                logger.debug('Cleanup[%s]: Triggering cleanup', self._task_id)
                self._on_cleanup(self)

    async def _mark_task_as_failed(self, exception: Exception) -> None:
        if self._exception is None:
            self._exception = exception
        if self._task_created.is_set():
            try:
                task = await self._task_manager.get_task()
                if task is not None:
                    await self._event_queue_agent.enqueue_event(
                        TaskStatusUpdateEvent(
                            task_id=task.id,
                            context_id=task.context_id,
                            status=TaskStatus(
                                state=TaskState.TASK_STATE_FAILED,
                            ),
                        )
                    )
            except QueueShutDown:
                pass

    async def get_task(self) -> Task:
        """Get task from db."""
        # TODO: THERE IS ZERO CONCURRENCY SAFETY HERE (Except inital task creation).
        await self._task_created.wait()
        task = await self._task_manager.get_task()
        if not task:
            raise RuntimeError('Task should have been created')
        return task



"""Helper functions for the A2A Python SDK."""

from a2a.helpers.agent_card import display_agent_card
from a2a.helpers.proto_helpers import (
    get_artifact_text,
    get_message_text,
    get_stream_response_text,
    get_text_parts,
    new_artifact,
    new_message,
    new_task,
    new_task_from_user_message,
    new_text_artifact,
    new_text_artifact_update_event,
    new_text_message,
    new_text_status_update_event,
)


__all__ = [
    'display_agent_card',
    'get_artifact_text',
    'get_message_text',
    'get_stream_response_text',
    'get_text_parts',
    'new_artifact',
    'new_message',
    'new_task',
    'new_task_from_user_message',
    'new_text_artifact',
    'new_text_artifact_update_event',
    'new_text_message',
    'new_text_status_update_event',
]



import logging

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Any

from google.protobuf.json_format import MessageToDict, Parse


if TYPE_CHECKING:
    from starlette.requests import Request

    from a2a.server.request_handlers.request_handler import RequestHandler

    _package_starlette_installed = True
else:
    try:
        from starlette.requests import Request

        _package_starlette_installed = True
    except ImportError:
        Request = Any

        _package_starlette_installed = False

from a2a.compat.v0_3 import a2a_v0_3_pb2 as pb2_v03
from a2a.compat.v0_3 import proto_utils
from a2a.compat.v0_3 import types as types_v03
from a2a.compat.v0_3.request_handler import RequestHandler03
from a2a.server.context import ServerCallContext
from a2a.utils import constants
from a2a.utils.telemetry import SpanKind, trace_class
from a2a.utils.version_validator import validate_version


logger = logging.getLogger(__name__)


@trace_class(kind=SpanKind.SERVER)
class REST03Handler:
    """Maps incoming REST-like (JSON+HTTP) requests to the appropriate request handler method and formats responses for v0.3 compatibility."""

    def __init__(
        self,
        request_handler: 'RequestHandler',
    ):
        """Initializes the REST03Handler.

        Args:
          request_handler: The underlying `RequestHandler` instance to delegate requests to (v1.0).
        """
        self.handler03 = RequestHandler03(request_handler=request_handler)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def on_message_send(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'message/send' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Returns:
            A `dict` containing the result (Task or Message) in v0.3 format.
        """
        body = await request.body()
        v03_pb_msg = pb2_v03.SendMessageRequest()
        Parse(body, v03_pb_msg, ignore_unknown_fields=True)
        v03_params_msg = proto_utils.FromProto.message_send_params(v03_pb_msg)
        rpc_req = types_v03.SendMessageRequest(id='', params=v03_params_msg)

        v03_resp = await self.handler03.on_message_send(rpc_req, context)

        pb2_v03_resp = proto_utils.ToProto.task_or_message(v03_resp)
        return MessageToDict(pb2_v03_resp)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def on_message_send_stream(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> AsyncIterator[dict[str, Any]]:
        """Handles the 'message/stream' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Yields:
            JSON serialized objects containing streaming events in v0.3 format.
        """
        body = await request.body()
        v03_pb_msg = pb2_v03.SendMessageRequest()
        Parse(body, v03_pb_msg, ignore_unknown_fields=True)
        v03_params_msg = proto_utils.FromProto.message_send_params(v03_pb_msg)
        rpc_req = types_v03.SendMessageRequest(id='', params=v03_params_msg)

        async for v03_stream_resp in self.handler03.on_message_send_stream(
            rpc_req, context
        ):
            v03_pb_resp = proto_utils.ToProto.stream_response(
                v03_stream_resp.result
            )
            yield MessageToDict(v03_pb_resp)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def on_cancel_task(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'tasks/cancel' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Returns:
            A `dict` containing the updated Task in v0.3 format.
        """
        task_id = request.path_params['id']
        rpc_req = types_v03.CancelTaskRequest(
            id='',
            params=types_v03.TaskIdParams(id=task_id),
        )

        v03_resp = await self.handler03.on_cancel_task(rpc_req, context)
        pb2_v03_task = proto_utils.ToProto.task(v03_resp)
        return MessageToDict(pb2_v03_task)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def on_subscribe_to_task(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> AsyncIterator[dict[str, Any]]:
        """Handles the 'tasks/{id}:subscribe' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Yields:
            JSON serialized objects containing streaming events in v0.3 format.
        """
        task_id = request.path_params['id']
        rpc_req = types_v03.TaskResubscriptionRequest(
            id='',
            params=types_v03.TaskIdParams(id=task_id),
        )

        async for v03_stream_resp in self.handler03.on_subscribe_to_task(
            rpc_req, context
        ):
            v03_pb_resp = proto_utils.ToProto.stream_response(
                v03_stream_resp.result
            )
            yield MessageToDict(v03_pb_resp)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def get_push_notification(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'tasks/pushNotificationConfig/get' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Returns:
            A `dict` containing the config in v0.3 format.
        """
        task_id = request.path_params['id']
        push_id = request.path_params['push_id']

        rpc_req = types_v03.GetTaskPushNotificationConfigRequest(
            id='',
            params=types_v03.GetTaskPushNotificationConfigParams(
                id=task_id, push_notification_config_id=push_id
            ),
        )

        v03_resp = await self.handler03.on_get_task_push_notification_config(
            rpc_req, context
        )
        pb2_v03_config = proto_utils.ToProto.task_push_notification_config(
            v03_resp
        )
        return MessageToDict(pb2_v03_config)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def set_push_notification(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'tasks/pushNotificationConfig/set' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Returns:
            A `dict` containing the config object in v0.3 format.
        """
        task_id = request.path_params['id']
        body = await request.body()

        v03_pb_push = pb2_v03.CreateTaskPushNotificationConfigRequest()
        Parse(body, v03_pb_push, ignore_unknown_fields=True)

        v03_params_push = (
            proto_utils.FromProto.task_push_notification_config_request(
                v03_pb_push
            )
        )
        v03_params_push.task_id = task_id

        rpc_req_push = types_v03.SetTaskPushNotificationConfigRequest(
            id='',
            params=v03_params_push,
        )

        v03_resp = await self.handler03.on_create_task_push_notification_config(
            rpc_req_push, context
        )
        pb2_v03_config = proto_utils.ToProto.task_push_notification_config(
            v03_resp
        )
        return MessageToDict(pb2_v03_config)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def on_get_task(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'v1/tasks/{id}' REST method.

        Args:
            request: The incoming `Request` object.
            context: Context provided by the server.

        Returns:
            A `Task` object containing the Task in v0.3 format.
        """
        task_id = request.path_params['id']
        history_length_str = request.query_params.get('historyLength')
        history_length = int(history_length_str) if history_length_str else None

        rpc_req = types_v03.GetTaskRequest(
            id='',
            params=types_v03.TaskQueryParams(
                id=task_id, history_length=history_length
            ),
        )

        v03_resp = await self.handler03.on_get_task(rpc_req, context)
        pb2_v03_task = proto_utils.ToProto.task(v03_resp)
        return MessageToDict(pb2_v03_task)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def list_push_notifications(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'tasks/pushNotificationConfig/list' REST method."""
        task_id = request.path_params['id']

        rpc_req = types_v03.ListTaskPushNotificationConfigRequest(
            id='',
            params=types_v03.ListTaskPushNotificationConfigParams(id=task_id),
        )

        v03_resp = await self.handler03.on_list_task_push_notification_configs(
            rpc_req, context
        )

        pb2_v03_resp = pb2_v03.ListTaskPushNotificationConfigResponse(
            configs=[
                proto_utils.ToProto.task_push_notification_config(c)
                for c in v03_resp
            ]
        )

        return MessageToDict(pb2_v03_resp)

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def list_tasks(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'tasks/list' REST method."""
        raise NotImplementedError('list tasks not implemented')

    @validate_version(constants.PROTOCOL_VERSION_0_3)
    async def on_get_extended_agent_card(
        self,
        request: Request,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        """Handles the 'v1/agent/authenticatedExtendedAgentCard' REST method."""
        rpc_req = types_v03.GetAuthenticatedExtendedCardRequest(id=0)
        v03_resp = await self.handler03.on_get_extended_agent_card(
            rpc_req, context
        )
        return v03_resp.model_dump(mode='json', exclude_none=True)



"""JSON-RPC application for A2A server."""

import json
import logging
import traceback

from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING, Any

from google.protobuf.json_format import MessageToDict, ParseDict
from jsonrpc.jsonrpc2 import JSONRPC20Request, JSONRPC20Response

from a2a.compat.v0_3.jsonrpc_adapter import JSONRPC03Adapter
from a2a.server.context import ServerCallContext
from a2a.server.events import Event
from a2a.server.jsonrpc_models import (
    InternalError,
    InvalidParamsError,
    InvalidRequestError,
    JSONParseError,
    JSONRPCError,
    MethodNotFoundError,
)
from a2a.server.request_handlers.request_handler import RequestHandler
from a2a.server.request_handlers.response_helpers import (
    build_error_response,
)
from a2a.server.routes.common import (
    DefaultServerCallContextBuilder,
    ServerCallContextBuilder,
)
from a2a.types.a2a_pb2 import (
    CancelTaskRequest,
    DeleteTaskPushNotificationConfigRequest,
    GetExtendedAgentCardRequest,
    GetTaskPushNotificationConfigRequest,
    GetTaskRequest,
    ListTaskPushNotificationConfigsRequest,
    ListTasksRequest,
    SendMessageRequest,
    SendMessageResponse,
    SubscribeToTaskRequest,
    Task,
    TaskPushNotificationConfig,
)
from a2a.utils import constants, proto_utils
from a2a.utils.errors import (
    A2AError,
    TaskNotFoundError,
    UnsupportedOperationError,
)
from a2a.utils.telemetry import SpanKind, trace_class
from a2a.utils.version_validator import validate_version


INTERNAL_ERROR_CODE = -32603

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from sse_starlette.sse import EventSourceResponse
    from starlette.exceptions import HTTPException
    from starlette.requests import Request
    from starlette.responses import JSONResponse, Response

    try:
        # Starlette v0.48.0
        from starlette.status import HTTP_413_CONTENT_TOO_LARGE
    except ImportError:
        from starlette.status import (  # type: ignore[no-redef]
            HTTP_413_REQUEST_ENTITY_TOO_LARGE as HTTP_413_CONTENT_TOO_LARGE,
        )

    _package_starlette_installed = True
else:
    try:
        from sse_starlette.sse import EventSourceResponse
        from starlette.exceptions import HTTPException
        from starlette.requests import Request
        from starlette.responses import JSONResponse, Response

        try:
            # Starlette v0.48.0
            from starlette.status import HTTP_413_CONTENT_TOO_LARGE
        except ImportError:
            from starlette.status import (
                HTTP_413_REQUEST_ENTITY_TOO_LARGE as HTTP_413_CONTENT_TOO_LARGE,
            )

        _package_starlette_installed = True
    except ImportError:
        _package_starlette_installed = False
        # Provide placeholder types for runtime type hinting when dependencies are not installed.
        # These will not be used if the code path that needs them is guarded by _http_server_installed.
        EventSourceResponse = Any
        HTTPException = Any
        Request = Any
        JSONResponse = Any
        Response = Any
        HTTP_413_CONTENT_TOO_LARGE = Any


@trace_class(kind=SpanKind.SERVER)
class JsonRpcDispatcher:
    """Base class for A2A JSONRPC applications.

    Handles incoming JSON-RPC requests, routes them to the appropriate
    handler methods, and manages response generation including Server-Sent Events
    (SSE).
    """

    # Method-to-model mapping for centralized routing
    # Proto types don't have model_fields, so we define the mapping explicitly
    # Method names match gRPC service method names
    METHOD_TO_MODEL: dict[str, type] = {
        'SendMessage': SendMessageRequest,
        'SendStreamingMessage': SendMessageRequest,  # Same proto type as SendMessage
        'GetTask': GetTaskRequest,
        'ListTasks': ListTasksRequest,
        'CancelTask': CancelTaskRequest,
        'CreateTaskPushNotificationConfig': TaskPushNotificationConfig,
        'GetTaskPushNotificationConfig': GetTaskPushNotificationConfigRequest,
        'ListTaskPushNotificationConfigs': ListTaskPushNotificationConfigsRequest,
        'DeleteTaskPushNotificationConfig': DeleteTaskPushNotificationConfigRequest,
        'SubscribeToTask': SubscribeToTaskRequest,
        'GetExtendedAgentCard': GetExtendedAgentCardRequest,
    }

    def __init__(
        self,
        request_handler: RequestHandler,
        context_builder: ServerCallContextBuilder | None = None,
        enable_v0_3_compat: bool = False,
    ) -> None:
        """Initializes the JsonRpcDispatcher.

        Args:
            request_handler: The handler instance responsible for processing A2A
              requests via http.
            context_builder: The ServerCallContextBuilder used to construct the
              ServerCallContext passed to the request_handler. If None the
              DefaultServerCallContextBuilder is used.
            enable_v0_3_compat: Whether to enable v0.3 backward compatibility on the same endpoint.
        """
        if not _package_starlette_installed:
            raise ImportError(
                'Packages `starlette` and `sse-starlette` are required to use the'
                ' `JsonRpcDispatcher`. They can be added as a part of `a2a-sdk`'
                ' optional dependencies, `a2a-sdk[http-server]`.'
            )

        self.request_handler = request_handler
        self._context_builder = (
            context_builder or DefaultServerCallContextBuilder()
        )
        self.enable_v0_3_compat = enable_v0_3_compat
        self._v03_adapter: JSONRPC03Adapter | None = None

        if self.enable_v0_3_compat:
            self._v03_adapter = JSONRPC03Adapter(
                http_handler=request_handler,
                context_builder=self._context_builder,
            )

    def _generate_error_response(
        self,
        request_id: str | int | None,
        error: Exception | JSONRPCError | A2AError,
    ) -> JSONResponse:
        """Creates a Starlette JSONResponse for a JSON-RPC error.

        Logs the error based on its type.

        Args:
            request_id: The ID of the request that caused the error.
            error: The error object (one of the JSONRPCError types).

        Returns:
            A `JSONResponse` object formatted as a JSON-RPC error response.
        """
        if not isinstance(error, A2AError | JSONRPCError):
            error = InternalError(message=str(error))

        response_data = build_error_response(request_id, error)
        error_info = response_data.get('error', {})
        code = error_info.get('code')
        message = error_info.get('message')
        data = error_info.get('data')

        log_level = logging.WARNING
        if code == INTERNAL_ERROR_CODE:
            log_level = logging.ERROR

        logger.log(
            log_level,
            "Request Error (ID: %s): Code=%s, Message='%s'%s",
            request_id,
            code,
            message,
            f', Data={data}' if data else '',
        )
        return JSONResponse(
            response_data,
            status_code=200,
        )

    async def handle_requests(self, request: Request) -> Response:  # noqa: PLR0911, PLR0912
        """Handles incoming POST requests to the main A2A endpoint.

        Parses the request body as JSON, validates it against A2A request types,
        dispatches it to the appropriate handler method, and returns the response.
        Handles JSON parsing errors, validation errors, and other exceptions,
        returning appropriate JSON-RPC error responses.

        Args:
            request: The incoming Starlette Request object.

        Returns:
            A Starlette Response object (JSONResponse or EventSourceResponse).

        Raises:
            (Implicitly handled): Various exceptions are caught and converted
            into JSON-RPC error responses by this method.
        """
        request_id = None
        body = None

        try:
            body = await request.json()
            if isinstance(body, dict):
                request_id = body.get('id')
                # Ensure request_id is valid for JSON-RPC response (str/int/None only)
                if request_id is not None and not isinstance(
                    request_id, str | int
                ):
                    request_id = None
            logger.debug('Request body: %s', body)
            # 1) Validate base JSON-RPC structure only (-32600 on failure)
            try:
                base_request = JSONRPC20Request.from_data(body)
                if not isinstance(base_request, JSONRPC20Request):
                    # Batch requests are not supported
                    return self._generate_error_response(
                        request_id,
                        InvalidRequestError(
                            message='Batch requests are not supported'
                        ),
                    )
                if body.get('jsonrpc') != '2.0':
                    return self._generate_error_response(
                        request_id,
                        InvalidRequestError(
                            message="Invalid request: 'jsonrpc' must be exactly '2.0'"
                        ),
                    )
            except Exception as e:
                logger.exception('Failed to validate base JSON-RPC request')
                return self._generate_error_response(
                    request_id,
                    InvalidRequestError(data=str(e)),
                )

            # 2) Route by method name; unknown -> -32601, known -> validate params (-32602 on failure)
            method: str | None = base_request.method
            request_id = base_request._id  # noqa: SLF001

            if not method:
                return self._generate_error_response(
                    request_id,
                    InvalidRequestError(message='Method is required'),
                )

            if (
                self.enable_v0_3_compat
                and self._v03_adapter
                and self._v03_adapter.supports_method(method)
            ):
                return await self._v03_adapter.handle_request(
                    request_id=request_id,
                    method=method,
                    body=body,
                    request=request,
                )

            model_class = self.METHOD_TO_MODEL.get(method)
            if not model_class:
                return self._generate_error_response(
                    request_id, MethodNotFoundError()
                )
            try:
                # Parse the params field into the proto message type
                params = body.get('params', {})
                specific_request = ParseDict(params, model_class())
            except Exception as e:
                logger.exception('Failed to parse request params')
                return self._generate_error_response(
                    request_id,
                    InvalidParamsError(data=str(e)),
                )

            # 3) Build call context and wrap the request for downstream handling
            call_context = self._context_builder.build(request)
            call_context.tenant = getattr(specific_request, 'tenant', '')
            call_context.state['method'] = method
            call_context.state['request_id'] = request_id

            # Route streaming requests by method name
            handler_result: (
                AsyncGenerator[dict[str, Any], None] | dict[str, Any]
            )
            if method in ('SendStreamingMessage', 'SubscribeToTask'):
                handler_result = await self._process_streaming_request(
                    request_id, specific_request, call_context
                )
            else:
                try:
                    raw_result = await self._process_non_streaming_request(
                        specific_request, call_context
                    )
                    handler_result = JSONRPC20Response(
                        result=raw_result, _id=request_id
                    ).data
                except A2AError as e:
                    handler_result = build_error_response(request_id, e)
            return self._create_response(call_context, handler_result)
        except json.decoder.JSONDecodeError as e:
            traceback.print_exc()
            return self._generate_error_response(
                None, JSONParseError(message=str(e))
            )
        except HTTPException as e:
            if e.status_code == HTTP_413_CONTENT_TOO_LARGE:
                return self._generate_error_response(
                    request_id,
                    InvalidRequestError(message='Payload too large'),
                )
            raise e
        except A2AError as e:
            return self._generate_error_response(request_id, e)
        except Exception as e:
            logger.exception('Unhandled exception')
            return self._generate_error_response(
                request_id, InternalError(message=str(e))
            )

    @validate_version(constants.PROTOCOL_VERSION_1_0)
    async def _process_streaming_request(
        self,
        request_id: str | int | None,
        request_obj: Any,
        context: ServerCallContext,
    ) -> AsyncGenerator[dict[str, Any], None]:
        """Processes streaming requests (SendStreamingMessage or SubscribeToTask).

        Args:
            request_id: The ID of the request.
            request_obj: The proto request message.
            context: The ServerCallContext for the request.

        Returns:
            An `AsyncGenerator` object to stream results to the client.
        """
        stream: AsyncGenerator | None = None
        method = context.state.get('method')
        if method == 'SendStreamingMessage':
            stream = self.request_handler.on_message_send_stream(
                request_obj, context
            )
        elif method == 'SubscribeToTask':
            stream = self.request_handler.on_subscribe_to_task(
                request_obj, context
            )

        if stream is None:
            raise UnsupportedOperationError(message='Stream not supported')

        # Eagerly fetch the first event to trigger validation/upfront errors
        try:
            first_event = await anext(stream)
        except StopAsyncIteration:
            first_event = None

        async def _wrap_stream(
            st: AsyncGenerator, first_evt: Event | None
        ) -> AsyncGenerator[dict[str, Any], None]:
            def _map_event(evt: Event) -> dict[str, Any]:
                stream_response = proto_utils.to_stream_response(evt)
                result = MessageToDict(
                    stream_response, preserving_proto_field_name=False
                )
                return JSONRPC20Response(result=result, _id=request_id).data

            try:
                if first_evt is not None:
                    yield _map_event(first_evt)

                async for event in st:
                    yield _map_event(event)
            except A2AError as e:
                yield build_error_response(request_id, e)

        return _wrap_stream(stream, first_event)

    async def _handle_send_message(
        self, request_obj: SendMessageRequest, context: ServerCallContext
    ) -> dict[str, Any]:
        task_or_message = await self.request_handler.on_message_send(
            request_obj, context
        )
        if isinstance(task_or_message, Task):
            return MessageToDict(SendMessageResponse(task=task_or_message))
        return MessageToDict(SendMessageResponse(message=task_or_message))

    async def _handle_cancel_task(
        self, request_obj: CancelTaskRequest, context: ServerCallContext
    ) -> dict[str, Any]:
        task = await self.request_handler.on_cancel_task(request_obj, context)
        if task:
            return MessageToDict(task, preserving_proto_field_name=False)
        raise TaskNotFoundError

    async def _handle_get_task(
        self, request_obj: GetTaskRequest, context: ServerCallContext
    ) -> dict[str, Any]:
        task = await self.request_handler.on_get_task(request_obj, context)
        if task:
            return MessageToDict(task, preserving_proto_field_name=False)
        raise TaskNotFoundError

    async def _handle_list_tasks(
        self, request_obj: ListTasksRequest, context: ServerCallContext
    ) -> dict[str, Any]:
        tasks_response = await self.request_handler.on_list_tasks(
            request_obj, context
        )
        return MessageToDict(
            tasks_response,
            preserving_proto_field_name=False,
            always_print_fields_with_no_presence=True,
        )

    async def _handle_create_task_push_notification_config(
        self,
        request_obj: TaskPushNotificationConfig,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        result_config = (
            await self.request_handler.on_create_task_push_notification_config(
                request_obj, context
            )
        )
        return MessageToDict(result_config, preserving_proto_field_name=False)

    async def _handle_get_task_push_notification_config(
        self,
        request_obj: GetTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        config = (
            await self.request_handler.on_get_task_push_notification_config(
                request_obj, context
            )
        )
        return MessageToDict(config, preserving_proto_field_name=False)

    async def _handle_list_task_push_notification_configs(
        self,
        request_obj: ListTaskPushNotificationConfigsRequest,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        configs_response = (
            await self.request_handler.on_list_task_push_notification_configs(
                request_obj, context
            )
        )
        return MessageToDict(
            configs_response, preserving_proto_field_name=False
        )

    async def _handle_delete_task_push_notification_config(
        self,
        request_obj: DeleteTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> None:
        await self.request_handler.on_delete_task_push_notification_config(
            request_obj, context
        )

    async def _handle_get_extended_agent_card(
        self,
        request_obj: GetExtendedAgentCardRequest,
        context: ServerCallContext,
    ) -> dict[str, Any]:
        card = await self.request_handler.on_get_extended_agent_card(
            request_obj, context
        )
        return MessageToDict(card, preserving_proto_field_name=False)

    @validate_version(constants.PROTOCOL_VERSION_1_0)
    async def _process_non_streaming_request(  # noqa: PLR0911
        self,
        request_obj: Any,
        context: ServerCallContext,
    ) -> dict[str, Any] | None:
        """Processes non-streaming requests.

        Args:
            request_obj: The proto request message.
            context: The ServerCallContext for the request.

        Returns:
            A dict containing the result or error.
        """
        method = context.state.get('method')
        match method:
            case 'SendMessage':
                return await self._handle_send_message(request_obj, context)
            case 'CancelTask':
                return await self._handle_cancel_task(request_obj, context)
            case 'GetTask':
                return await self._handle_get_task(request_obj, context)
            case 'ListTasks':
                return await self._handle_list_tasks(request_obj, context)
            case 'CreateTaskPushNotificationConfig':
                return await self._handle_create_task_push_notification_config(
                    request_obj, context
                )
            case 'GetTaskPushNotificationConfig':
                return await self._handle_get_task_push_notification_config(
                    request_obj, context
                )
            case 'ListTaskPushNotificationConfigs':
                return await self._handle_list_task_push_notification_configs(
                    request_obj, context
                )
            case 'DeleteTaskPushNotificationConfig':
                await self._handle_delete_task_push_notification_config(
                    request_obj, context
                )
                return None
            case 'GetExtendedAgentCard':
                return await self._handle_get_extended_agent_card(
                    request_obj, context
                )
            case _:
                logger.error('Unhandled method: %s', method)
                raise UnsupportedOperationError(
                    message=f'Method {method} is not supported.'
                )

    def _create_response(
        self,
        context: ServerCallContext,
        handler_result: AsyncGenerator[dict[str, Any]] | dict[str, Any],
    ) -> Response:
        """Creates a Starlette Response based on the result from the request handler.

        Handles:
        - AsyncGenerator for Server-Sent Events (SSE).
        - Dict responses from handlers.

        Args:
            context: The ServerCallContext provided to the request handler.
            handler_result: The result from a request handler method. Can be an
                async generator for streaming or a dict for non-streaming.

        Returns:
            A Starlette JSONResponse or EventSourceResponse.
        """
        if isinstance(handler_result, AsyncGenerator):
            # Result is a stream of dict objects
            async def event_generator(
                stream: AsyncGenerator[dict[str, Any]],
            ) -> AsyncGenerator[dict[str, str]]:
                try:
                    async for item in stream:
                        event: dict[str, str] = {
                            'data': json.dumps(item),
                        }
                        if 'error' in item:
                            event['event'] = 'error'
                        yield event
                except Exception as e:
                    logger.exception(
                        'Unhandled error during JSON-RPC SSE stream'
                    )
                    rpc_error: A2AError | JSONRPCError = (
                        e
                        if isinstance(e, A2AError | JSONRPCError)
                        else InternalError(message=str(e))
                    )
                    error_response = build_error_response(
                        context.state.get('request_id'), rpc_error
                    )
                    yield {
                        'event': 'error',
                        'data': json.dumps(error_response),
                    }

            return EventSourceResponse(event_generator(handler_result))

        # handler_result is a dict (JSON-RPC response)
        return JSONResponse(handler_result)



import logging

from collections.abc import AsyncGenerator, Callable
from functools import wraps
from typing import Any, NoReturn

from a2a.client.errors import A2AClientError, A2AClientTimeoutError
from a2a.utils.errors import JSON_RPC_ERROR_CODE_MAP


try:
    import grpc  # type: ignore[reportMissingModuleSource]
except ImportError as e:
    raise ImportError(
        'A2AGrpcClient requires grpcio and grpcio-tools to be installed. '
        'Install with: '
        "'pip install a2a-sdk[grpc]'"
    ) from e


from a2a.client.client import ClientCallContext, ClientConfig
from a2a.client.optionals import Channel
from a2a.client.transports.base import ClientTransport
from a2a.compat.v0_3 import (
    a2a_v0_3_pb2,
    a2a_v0_3_pb2_grpc,
    conversions,
    proto_utils,
)
from a2a.compat.v0_3 import (
    types as types_v03,
)
from a2a.compat.v0_3.extension_headers import add_legacy_extension_header
from a2a.types import a2a_pb2
from a2a.utils.constants import PROTOCOL_VERSION_0_3, VERSION_HEADER
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)

_A2A_ERROR_NAME_TO_CLS = {
    error_type.__name__: error_type for error_type in JSON_RPC_ERROR_CODE_MAP
}


def _map_grpc_error(e: grpc.aio.AioRpcError) -> NoReturn:
    if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
        raise A2AClientTimeoutError('Client Request timed out') from e

    details = e.details()
    if isinstance(details, str) and ': ' in details:
        error_type_name, error_message = details.split(': ', 1)
        exception_cls = _A2A_ERROR_NAME_TO_CLS.get(error_type_name)
        if exception_cls:
            raise exception_cls(error_message) from e
    raise A2AClientError(f'gRPC Error {e.code().name}: {e.details()}') from e


def _handle_grpc_exception(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return await func(*args, **kwargs)
        except grpc.aio.AioRpcError as e:
            _map_grpc_error(e)

    return wrapper


def _handle_grpc_stream_exception(
    func: Callable[..., Any],
) -> Callable[..., Any]:
    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            async for item in func(*args, **kwargs):
                yield item
        except grpc.aio.AioRpcError as e:
            _map_grpc_error(e)

    return wrapper


@trace_class(kind=SpanKind.CLIENT)
class CompatGrpcTransport(ClientTransport):
    """A backward compatible gRPC transport for A2A v0.3."""

    def __init__(self, channel: Channel, agent_card: a2a_pb2.AgentCard | None):
        """Initializes the CompatGrpcTransport."""
        self.agent_card = agent_card
        self.channel = channel
        self.stub = a2a_v0_3_pb2_grpc.A2AServiceStub(channel)

    @classmethod
    def create(
        cls,
        card: a2a_pb2.AgentCard,
        url: str,
        config: ClientConfig,
    ) -> 'CompatGrpcTransport':
        """Creates a gRPC transport for the A2A client."""
        if config.grpc_channel_factory is None:
            raise ValueError('grpc_channel_factory is required when using gRPC')
        return cls(config.grpc_channel_factory(url), card)

    @_handle_grpc_exception
    async def send_message(
        self,
        request: a2a_pb2.SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.SendMessageResponse:
        """Sends a non-streaming message request to the agent (v0.3)."""
        req_v03 = conversions.to_compat_send_message_request(
            request, request_id=0
        )
        req_proto = a2a_v0_3_pb2.SendMessageRequest(
            request=proto_utils.ToProto.message(req_v03.params.message),
            configuration=proto_utils.ToProto.message_send_configuration(
                req_v03.params.configuration
            ),
            metadata=proto_utils.ToProto.metadata(req_v03.params.metadata),
        )

        resp_proto = await self.stub.SendMessage(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )

        which = resp_proto.WhichOneof('payload')
        if which == 'task':
            return a2a_pb2.SendMessageResponse(
                task=conversions.to_core_task(
                    proto_utils.FromProto.task(resp_proto.task)
                )
            )
        if which == 'msg':
            return a2a_pb2.SendMessageResponse(
                message=conversions.to_core_message(
                    proto_utils.FromProto.message(resp_proto.msg)
                )
            )
        return a2a_pb2.SendMessageResponse()

    @_handle_grpc_stream_exception
    async def send_message_streaming(
        self,
        request: a2a_pb2.SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[a2a_pb2.StreamResponse]:
        """Sends a streaming message request to the agent (v0.3)."""
        req_v03 = conversions.to_compat_send_message_request(
            request, request_id=0
        )
        req_proto = a2a_v0_3_pb2.SendMessageRequest(
            request=proto_utils.ToProto.message(req_v03.params.message),
            configuration=proto_utils.ToProto.message_send_configuration(
                req_v03.params.configuration
            ),
            metadata=proto_utils.ToProto.metadata(req_v03.params.metadata),
        )

        stream = self.stub.SendStreamingMessage(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        while True:
            response = await stream.read()
            if response == grpc.aio.EOF:  # type: ignore[attr-defined]
                break
            yield conversions.to_core_stream_response(
                types_v03.SendStreamingMessageSuccessResponse(
                    result=proto_utils.FromProto.stream_response(response)
                )
            )

    @_handle_grpc_stream_exception
    async def subscribe(
        self,
        request: a2a_pb2.SubscribeToTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[a2a_pb2.StreamResponse]:
        """Reconnects to get task updates (v0.3)."""
        req_proto = a2a_v0_3_pb2.TaskSubscriptionRequest(
            name=f'tasks/{request.id}'
        )

        stream = self.stub.TaskSubscription(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        while True:
            response = await stream.read()
            if response == grpc.aio.EOF:  # type: ignore[attr-defined]
                break
            yield conversions.to_core_stream_response(
                types_v03.SendStreamingMessageSuccessResponse(
                    result=proto_utils.FromProto.stream_response(response)
                )
            )

    @_handle_grpc_exception
    async def get_task(
        self,
        request: a2a_pb2.GetTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.Task:
        """Retrieves the current state and history of a specific task (v0.3)."""
        req_proto = a2a_v0_3_pb2.GetTaskRequest(
            name=f'tasks/{request.id}',
            history_length=request.history_length,
        )
        resp_proto = await self.stub.GetTask(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        return conversions.to_core_task(proto_utils.FromProto.task(resp_proto))

    @_handle_grpc_exception
    async def list_tasks(
        self,
        request: a2a_pb2.ListTasksRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.ListTasksResponse:
        """Retrieves tasks for an agent (v0.3 - NOT SUPPORTED in v0.3)."""
        # v0.3 proto doesn't have ListTasks.
        raise NotImplementedError(
            'ListTasks is not supported in A2A v0.3 gRPC.'
        )

    @_handle_grpc_exception
    async def cancel_task(
        self,
        request: a2a_pb2.CancelTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.Task:
        """Requests the agent to cancel a specific task (v0.3)."""
        req_proto = a2a_v0_3_pb2.CancelTaskRequest(name=f'tasks/{request.id}')
        resp_proto = await self.stub.CancelTask(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        return conversions.to_core_task(proto_utils.FromProto.task(resp_proto))

    @_handle_grpc_exception
    async def create_task_push_notification_config(
        self,
        request: a2a_pb2.TaskPushNotificationConfig,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.TaskPushNotificationConfig:
        """Sets or updates the push notification configuration (v0.3)."""
        req_v03 = (
            conversions.to_compat_create_task_push_notification_config_request(
                request, request_id=0
            )
        )
        req_proto = a2a_v0_3_pb2.CreateTaskPushNotificationConfigRequest(
            parent=f'tasks/{request.task_id}',
            config_id=req_v03.params.push_notification_config.id,
            config=proto_utils.ToProto.task_push_notification_config(
                req_v03.params
            ),
        )
        resp_proto = await self.stub.CreateTaskPushNotificationConfig(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        return conversions.to_core_task_push_notification_config(
            proto_utils.FromProto.task_push_notification_config(resp_proto)
        )

    @_handle_grpc_exception
    async def get_task_push_notification_config(
        self,
        request: a2a_pb2.GetTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.TaskPushNotificationConfig:
        """Retrieves the push notification configuration (v0.3)."""
        req_proto = a2a_v0_3_pb2.GetTaskPushNotificationConfigRequest(
            name=f'tasks/{request.task_id}/pushNotificationConfigs/{request.id}'
        )
        resp_proto = await self.stub.GetTaskPushNotificationConfig(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        return conversions.to_core_task_push_notification_config(
            proto_utils.FromProto.task_push_notification_config(resp_proto)
        )

    @_handle_grpc_exception
    async def list_task_push_notification_configs(
        self,
        request: a2a_pb2.ListTaskPushNotificationConfigsRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.ListTaskPushNotificationConfigsResponse:
        """Lists push notification configurations for a specific task (v0.3)."""
        req_proto = a2a_v0_3_pb2.ListTaskPushNotificationConfigRequest(
            parent=f'tasks/{request.task_id}'
        )
        resp_proto = await self.stub.ListTaskPushNotificationConfig(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        return conversions.to_core_list_task_push_notification_config_response(
            proto_utils.FromProto.list_task_push_notification_config_response(
                resp_proto
            )
        )

    @_handle_grpc_exception
    async def delete_task_push_notification_config(
        self,
        request: a2a_pb2.DeleteTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> None:
        """Deletes the push notification configuration (v0.3)."""
        req_proto = a2a_v0_3_pb2.DeleteTaskPushNotificationConfigRequest(
            name=f'tasks/{request.task_id}/pushNotificationConfigs/{request.id}'
        )
        await self.stub.DeleteTaskPushNotificationConfig(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )

    @_handle_grpc_exception
    async def get_extended_agent_card(
        self,
        request: a2a_pb2.GetExtendedAgentCardRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> a2a_pb2.AgentCard:
        """Retrieves the agent's card (v0.3)."""
        req_proto = a2a_v0_3_pb2.GetAgentCardRequest()
        resp_proto = await self.stub.GetAgentCard(
            req_proto,
            metadata=self._get_grpc_metadata(context),
        )
        card = conversions.to_core_agent_card(
            proto_utils.FromProto.agent_card(resp_proto)
        )

        self.agent_card = card
        return card

    async def close(self) -> None:
        """Closes the gRPC channel."""
        await self.channel.close()

    def _get_grpc_metadata(
        self, context: ClientCallContext | None = None
    ) -> list[tuple[str, str]]:
        """Creates gRPC metadata for extensions."""
        metadata = [(VERSION_HEADER.lower(), PROTOCOL_VERSION_0_3)]

        if context and context.service_parameters:
            params = dict(context.service_parameters)
            add_legacy_extension_header(params)
            for key, value in params.items():
                metadata.append((key.lower(), value))

        return metadata



"""add_columns_owner_last_updated.

Revision ID: 6419d2d130f6
Revises:
Create Date: 2026-02-17 09:23:06.758085

"""

import logging
from collections.abc import Sequence

import sqlalchemy as sa

try:
    from alembic import context
except ImportError as e:
    raise ImportError(
        "'Add columns owner and last_updated to database tables' migration requires Alembic. Install with: 'pip install a2a-sdk[db-cli]'."
    ) from e

from a2a.migrations.migration_utils import (
    table_exists,
    add_column,
    add_index,
    drop_column,
    drop_index,
)


# revision identifiers, used by Alembic.
revision: str = '6419d2d130f6'
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # Get the default value from the config (passed via CLI)
    owner = context.config.get_main_option(
        'add_columns_owner_last_updated_default_owner',
        'legacy_v03_no_user_info',
    )
    tasks_table = context.config.get_main_option('tasks_table', 'tasks')
    push_notification_configs_table = context.config.get_main_option(
        'push_notification_configs_table', 'push_notification_configs'
    )

    if table_exists(tasks_table):
        add_column(tasks_table, 'owner', True, sa.String(255), owner)
        add_column(tasks_table, 'last_updated', True, sa.DateTime())
        add_index(
            tasks_table,
            f'idx_{tasks_table}_owner_last_updated',
            ['owner', 'last_updated'],
        )
    else:
        logging.warning(
            f"Table '{tasks_table}' does not exist. Skipping upgrade for this table."
        )

    if table_exists(push_notification_configs_table):
        add_column(
            push_notification_configs_table,
            'owner',
            True,
            sa.String(255),
            owner,
        )
        add_index(
            push_notification_configs_table,
            f'ix_{push_notification_configs_table}_owner',
            ['owner'],
        )
    else:
        logging.warning(
            f"Table '{push_notification_configs_table}' does not exist. Skipping upgrade for this table."
        )


def downgrade() -> None:
    """Downgrade schema."""
    tasks_table = context.config.get_main_option('tasks_table', 'tasks')
    push_notification_configs_table = context.config.get_main_option(
        'push_notification_configs_table', 'push_notification_configs'
    )

    if table_exists(tasks_table):
        drop_index(
            tasks_table,
            f'idx_{tasks_table}_owner_last_updated',
        )
        drop_column(tasks_table, 'owner')
        drop_column(tasks_table, 'last_updated')
    else:
        logging.warning(
            f"Table '{tasks_table}' does not exist. Skipping downgrade for this table."
        )

    if table_exists(push_notification_configs_table):
        drop_index(
            push_notification_configs_table,
            f'ix_{push_notification_configs_table}_owner',
        )
        drop_column(push_notification_configs_table, 'owner')
    else:
        logging.warning(
            f"Table '{push_notification_configs_table}' does not exist. Skipping downgrade for this table."
        )



from abc import ABC, abstractmethod

from a2a.types.a2a_pb2 import (
    Task,
    TaskArtifactUpdateEvent,
    TaskStatusUpdateEvent,
)


PushNotificationEvent = Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent


class PushNotificationSender(ABC):
    """Interface for sending push notifications for tasks."""

    @abstractmethod
    async def send_notification(
        self, task_id: str, event: PushNotificationEvent
    ) -> None:
        """Sends a push notification containing the latest task state."""



import logging

from typing import TYPE_CHECKING, Any

from a2a.compat.v0_3.rest_adapter import REST03Adapter
from a2a.server.request_handlers.request_handler import RequestHandler
from a2a.server.routes.common import ServerCallContextBuilder
from a2a.server.routes.rest_dispatcher import RestDispatcher


if TYPE_CHECKING:
    from starlette.routing import BaseRoute, Mount, Route

    _package_starlette_installed = True
else:
    try:
        from starlette.routing import BaseRoute, Mount, Route

        _package_starlette_installed = True
    except ImportError:
        Route = Any
        Mount = Any
        BaseRoute = Any

        _package_starlette_installed = False

logger = logging.getLogger(__name__)


def create_rest_routes(
    request_handler: RequestHandler,
    context_builder: ServerCallContextBuilder | None = None,
    enable_v0_3_compat: bool = False,
    path_prefix: str = '',
) -> list['BaseRoute']:
    """Creates the Starlette Routes for the A2A protocol REST endpoint.

    Args:
        request_handler: The handler instance responsible for processing A2A
          requests via http.
        context_builder: The ServerCallContextBuilder used to construct the
          ServerCallContext passed to the request_handler. If None the
          DefaultServerCallContextBuilder is used.
        enable_v0_3_compat: If True, mounts backward-compatible v0.3 protocol
          endpoints using REST03Adapter.
        path_prefix: The URL prefix for the REST endpoints.
    """
    if not _package_starlette_installed:
        raise ImportError(
            'Packages `starlette` and `sse-starlette` are required to use'
            ' the `create_rest_routes`. They can be added as a part of `a2a-sdk` '
            'optional dependencies, `a2a-sdk[http-server]`.'
        )

    dispatcher = RestDispatcher(
        request_handler=request_handler,
        context_builder=context_builder,
    )

    routes: list[BaseRoute] = []
    if enable_v0_3_compat:
        v03_adapter = REST03Adapter(
            http_handler=request_handler,
            context_builder=context_builder,
        )
        v03_routes = v03_adapter.routes()
        for (path, method), endpoint in v03_routes.items():
            routes.append(
                Route(
                    path=f'{path_prefix}{path}',
                    endpoint=endpoint,
                    methods=[method],
                )
            )

    base_routes = {
        ('/message:send', 'POST'): dispatcher.on_message_send,
        ('/message:stream', 'POST'): dispatcher.on_message_send_stream,
        ('/tasks/{id}:cancel', 'POST'): dispatcher.on_cancel_task,
        ('/tasks/{id}:subscribe', 'GET'): dispatcher.on_subscribe_to_task,
        ('/tasks/{id}:subscribe', 'POST'): dispatcher.on_subscribe_to_task,
        ('/tasks/{id}', 'GET'): dispatcher.on_get_task,
        (
            '/tasks/{id}/pushNotificationConfigs/{push_id}',
            'GET',
        ): dispatcher.get_push_notification,
        (
            '/tasks/{id}/pushNotificationConfigs/{push_id}',
            'DELETE',
        ): dispatcher.delete_push_notification,
        (
            '/tasks/{id}/pushNotificationConfigs',
            'POST',
        ): dispatcher.set_push_notification,
        (
            '/tasks/{id}/pushNotificationConfigs',
            'GET',
        ): dispatcher.list_push_notifications,
        ('/tasks', 'GET'): dispatcher.list_tasks,
        (
            '/extendedAgentCard',
            'GET',
        ): dispatcher.handle_authenticated_agent_card,
    }

    base_route_objects = []
    for (path, method), endpoint in base_routes.items():
        base_route_objects.append(
            Route(
                path=f'{path_prefix}{path}',
                endpoint=endpoint,
                methods=[method],
            )
        )
    routes.extend(base_route_objects)
    routes.append(Mount(path='/{tenant}', routes=base_route_objects))

    return routes
